#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import collections
import json
import math
import numbers
import os
import re
import sys

from paleomix.node import Node, NodeError
from paleomix.common.utilities import set_in, get_in
from paleomix.common.fileutils import move_file, reroot_path
from paleomix.tools.bam_stats.coverage import read_table as read_coverage_table
from paleomix.common.bedtools import BEDRecord
from paleomix.common.formats.fasta import FASTA

import paleomix.common.text as text


_PE_READS = frozenset(("Paired", "Singleton", "Collapsed", "CollapsedTruncated"))
_SE_READS = frozenset(("Single",))


class SummaryTableNode(Node):
    def __init__(
        self, config, makefile, target, cov_for_lanes, cov_for_libs, dependencies=()
    ):
        self._target = target.name
        self._output_file = os.path.join(config.destination, self._target + ".summary")
        self._prefixes = makefile["Prefixes"]

        self._in_raw_bams = cov_for_lanes
        self._in_lib_bams = cov_for_libs
        input_files = set()
        input_files.update(sum(map(list, self._in_raw_bams.values()), []))
        input_files.update(sum(map(list, self._in_lib_bams.values()), []))

        self._in_raw_read = collections.defaultdict(list)
        for prefix in target.prefixes:
            for sample in prefix.samples:
                for library in sample.libraries:
                    for lane in library.lanes:
                        filenames = []
                        filetype = None

                        if lane.reads.stats:
                            filetype = "Raw"
                            filenames.append(lane.reads.stats)
                        elif lane.reads.validation:
                            filenames.extend(lane.reads.validation)

                            fileset = set(lane.reads.files)
                            if fileset & _PE_READS and fileset & _SE_READS:
                                filetype = "*"
                            elif fileset & _PE_READS:
                                filetype = "PE"
                            elif fileset & _SE_READS:
                                filetype = "SE"
                            else:
                                assert False, lane.reads.files
                                continue
                        else:
                            assert False

                        input_files.update(filenames)
                        self._in_raw_read[(sample.name, library.name, lane.name)] = (
                            filetype,
                            filenames,
                        )

        Node.__init__(
            self,
            description="writing summary to %s" % (self._output_file,),
            input_files=[_f for _f in input_files if _f],
            output_files=[self._output_file],
            dependencies=dependencies,
        )

    def _run(self, config, temp):
        rois = self._stat_areas_of_interest(self._prefixes)
        genomes = self._stat_prefixes(self._prefixes)
        with open(reroot_path(temp, self._output_file), "w") as table:
            table.write("# Command:\n")
            table.write("#     %s\n" % (" ".join(sys.argv)))
            table.write("#\n")
            table.write("# Directory:\n")
            table.write("#     %s\n" % (os.getcwd()))
            table.write("#\n")
            self._write_genomes(table, genomes)
            table.write("#\n")
            self._write_areas_of_interest(table, rois)
            table.write("#\n#\n")

            for key, roi in rois.items():
                genomes[":".join(key)] = {"Size": roi["Size"]}
            self._write_tables(table, genomes)

    def _teardown(self, _config, temp):
        move_file(reroot_path(temp, self._output_file), self._output_file)

    def _write_genomes(self, table, genomes):
        table.write("# Genomes:\n")
        rows = [["Name", "Contigs", "Size", "Prefix"]]
        for (_, prefix) in sorted(self._prefixes.items()):
            stats = genomes[prefix["Name"]]
            rows.append(
                (
                    prefix["Name"],
                    stats["NContigs"],
                    stats["Size"],
                    prefix["Path"],
                )
            )

        for line in text.padded_table(rows):
            table.write("#     %s\n" % (line,))

    def _write_areas_of_interest(self, table, rois):
        table.write("# Regions Of Interest:\n")
        rows = [["Genome", "ROI", "Size", "NFeatures", "NIntervals", "Path"]]
        for (_, roi) in sorted(rois.items()):
            rows.append(
                [
                    roi[key]
                    for key in (
                        "Genome",
                        "Name",
                        "Size",
                        "NFeatures",
                        "NIntervals",
                        "Path",
                    )
                ]
            )

        for line in text.padded_table(rows):
            table.write("#     %s\n" % (line,))

    def _write_tables(self, out, genomes):
        for row in self._build_tables(genomes):
            out.write("%s\n" % "\t".join(map(str, row)))

    def _build_tables(self, genomes):
        yield ["Target", "Sample", "Library", "Measure", "Value", "# Description"]

        for (target, samples) in sorted(
            self._read_tables(self._prefixes, genomes).items()
        ):
            for (sample, libraries) in sorted(samples.items()):
                for (library, prefixes) in sorted(libraries.items()):
                    ordered = (
                        [("reads", prefixes.pop("reads"))]
                        if "reads" in prefixes
                        else []
                    )
                    ordered.extend(sorted(prefixes.items()))

                    for (prefix, table) in ordered:
                        table.pop("hits_unique_nts(%s)" % prefix, None)

                        for (key, value) in sorted(
                            table.items(), key=_measure_ordering
                        ):
                            if isinstance(value, numbers.Number) and math.isnan(value):
                                value = "NA"

                            yield (
                                target,
                                sample,
                                library,
                                key,
                                value,
                                _get_comment(key),
                            )
                        yield ""
                    yield ""

    def _read_tables(self, prefixes, genomes):
        table = {}
        self._read_reads_settings(table)
        self._read_raw_bam_stats(table)
        self._read_lib_bam_stats(table)

        for (target, samples) in table.items():
            merged_samples = {}
            for (sample, libraries) in samples.items():
                merged_libraries = {}
                for (library, subtables) in libraries.items():
                    for (tblname, subtable) in subtables.items():
                        merged_libraries[tblname] = self._merge_tables(
                            (merged_libraries.get(tblname, {}), subtable)
                        )
                        merged_samples[tblname] = self._merge_tables(
                            (merged_samples.get(tblname, {}), subtable)
                        )
                    libraries[library] = self._annotate_subtables(subtables, genomes)
                set_in(
                    table,
                    (target, sample, "*"),
                    self._annotate_subtables(merged_libraries, genomes),
                )
            set_in(
                table,
                (target, "*", "*"),
                self._annotate_subtables(merged_samples, genomes),
            )

        return table

    @classmethod
    def _annotate_subtables(cls, subtables, genomes):
        for (tblname, subtable) in subtables.items():
            if tblname == "reads":
                fractions = [
                    ("seq_trash_se", "seq_reads_se", "seq_trash_se_frac"),
                    ("seq_trash_pe_1", "seq_reads_pairs", "seq_trash_pe_1_frac"),
                    ("seq_trash_pe_2", "seq_reads_pairs", "seq_trash_pe_2_frac"),
                    ("seq_collapsed", "seq_reads_pairs", "seq_collapsed_frac"),
                    ("seq_retained_nts", "seq_retained_reads", "seq_retained_length"),
                ]

                for (numerator, denominator, measure) in fractions:
                    if (numerator in subtable) and (denominator in subtable):
                        value = subtable[numerator] / float(
                            subtable[denominator] or "NaN"
                        )
                        subtable[measure] = value
            else:
                total_hits = subtable["hits_raw(%s)" % tblname]
                total_nts = subtable["hits_unique_nts(%s)" % tblname]
                total_uniq = subtable["hits_unique(%s)" % tblname]
                total_reads = subtables.get("reads", {}).get("seq_retained_reads", 0)
                genome_size = genomes[tblname]["Size"]

                subtable["hits_raw_frac(%s)" % tblname] = total_hits / float(
                    total_reads or "NaN"
                )
                subtable["hits_unique_frac(%s)" % tblname] = total_uniq / float(
                    total_reads or "NaN"
                )
                subtable["hits_clonality(%s)" % tblname] = 1 - total_uniq / float(
                    total_hits or "NaN"
                )
                subtable["hits_length(%s)" % tblname] = total_nts / float(
                    total_uniq or "NaN"
                )
                subtable["hits_coverage(%s)" % tblname] = total_nts / float(
                    genome_size or "NaN"
                )

        return subtables

    def _read_reads_settings(self, table):
        for (
            (sample, library, barcode),
            (filetype, filenames),
        ) in self._in_raw_read.items():
            key = (self._target, sample, library, "reads", barcode)
            set_in(table, key, self._stat_read_settings(filetype, filenames))

        for (_, samples) in table.items():
            for (sample, libraries) in samples.items():
                for (library, prefixes) in libraries.items():
                    prefixes["reads"] = self._merge_tables(
                        list(prefixes["reads"].values())
                    )

        return table

    def _read_raw_bam_stats(self, table):
        for ((genome, target, sample, library), filenames) in self._in_raw_bams.items():
            key = (target, sample, library)
            hits, _ = self._read_coverage_tables(key, filenames)

            set_in(
                table, (target, sample, library, genome, "hits_raw(%s)" % genome), hits
            )

    def _read_lib_bam_stats(self, table):
        for ((genome, target, sample, library), filenames) in self._in_lib_bams.items():
            key = (target, sample, library)
            hits, nts = self._read_coverage_tables(key, filenames)

            set_in(
                table,
                (target, sample, library, genome, "hits_unique(%s)" % genome),
                hits,
            )
            set_in(
                table,
                (target, sample, library, genome, "hits_unique_nts(%s)" % genome),
                nts,
            )

    @classmethod
    def _read_coverage_tables(cls, key, filenames):
        hits = nts = 0
        for filename in filenames:
            subtable = {}
            read_coverage_table(subtable, filename)
            contigtables = get_in(subtable, key)

            if contigtables is None:
                raise NodeError(
                    "Error reading table %r; row not found:"
                    "\n   %s\n\nIf files have been renamed "
                    "during the run, then please remove this file "
                    "in that it may be re-generated.\nHowever, "
                    "note that read-group tags in the BAM files "
                    "may not be correct!" % (filename, "   ".join(key))
                )

            for contigtable in contigtables.values():
                hits += contigtable["Hits"]
                nts += contigtable["M"]
        return hits, nts

    @classmethod
    def _merge_tables(cls, tables):
        merged = {}
        for table in tables:
            for (measure, value) in table.items():
                if not isinstance(value, numbers.Number):
                    other = merged.get(measure, value)
                    merged[measure] = value if (value == other) else "*"
                else:
                    merged[measure] = value + merged.get(measure, 0)
        return merged

    @classmethod
    def _stat_read_settings(cls, filetype, filenames):
        assert filetype in ("Raw", "SE", "PE", "*", "BAM"), filetype

        if filetype != "Raw":
            nan = float("nan")
            stats = {
                "lib_type": filetype,
                "seq_retained_nts": 0,
                "seq_retained_reads": 0,
            }

            if filetype in ("SE", "*"):
                stats["seq_reads_se"] = nan
                stats["seq_trash_se"] = nan

            if filetype in ("PE", "*"):
                stats["seq_reads_pairs"] = nan
                stats["seq_trash_pe_1"] = nan
                stats["seq_trash_pe_2"] = nan
                stats["seq_collapsed"] = 0

            for filename in filenames:
                with open(filename) as handle:
                    for key, value in json.load(handle).items():
                        if key in stats:
                            stats[key] += int(value)

            return stats

        assert len(filenames) == 1, filenames
        with open(filenames[0]) as settings_file:
            settings = settings_file.read()

            def _re_search(regexp, default=None):
                match = re.search(regexp, settings)
                if not match:
                    if default is not None:
                        return default
                    raise KeyError(
                        "Could not find match with RegExp %s in file '%s'"
                        % (repr(regexp), filename)
                    )

                return int(match.groups()[0])

            if "Paired end mode" in settings or "paired-end reads" in settings:
                return {
                    "lib_type": "PE",
                    "seq_reads_pairs": _re_search("number of read pairs: ([0-9]+)"),
                    "seq_trash_pe_1": _re_search("discarded mate 1 reads: ([0-9]+)"),
                    "seq_trash_pe_2": _re_search("discarded mate 2 reads: ([0-9]+)"),
                    "seq_retained_nts": _re_search("retained nucleotides: ([0-9]+)"),
                    "seq_retained_reads": _re_search("retained reads: ([0-9]+)"),
                    "seq_collapsed": _re_search(
                        "of (?:full-length )?collapsed pairs: ([0-9]+)", 0
                    )
                    + _re_search("of truncated collapsed pairs: ([0-9]+)", 0),
                }
            elif "Single end mode" in settings or "single-end reads" in settings:
                return {
                    "lib_type": "SE",
                    "seq_reads_se": _re_search(
                        "number of (?:reads|read pairs): ([0-9]+)"
                    ),
                    "seq_trash_se": _re_search("discarded mate 1 reads: ([0-9]+)"),
                    "seq_retained_nts": _re_search("retained nucleotides: ([0-9]+)"),
                    "seq_retained_reads": _re_search("retained reads: ([0-9]+)"),
                }
            else:
                assert False, filename

    @classmethod
    def _stat_areas_of_interest(cls, prefixes):
        """Returns (size, number of named intervals, total number of intervals)
        for a set of areas of interest."""
        areas_of_interest = {}
        for (prefix_name, prefix) in prefixes.items():
            for (roi_name, roi_filename) in prefix.get("RegionsOfInterest", {}).items():
                count, names, size = 0, set(), 0
                with open(roi_filename) as handle:
                    for line in handle:
                        bed = BEDRecord(line)
                        names.add(bed.name if len(bed) >= 4 else (bed.contig + "*"))
                        size += bed.end - bed.start
                        count += 1
                areas_of_interest[(prefix_name, roi_name)] = {
                    "Size": size,
                    "NFeatures": len(names),
                    "NIntervals": count,
                    "Genome": prefix["Name"],
                    "Name": roi_name,
                    "Path": roi_filename,
                }
        return areas_of_interest

    @classmethod
    def _stat_prefixes(cls, prefixes):
        """Returns (size, number of contigs) for a set of BWA prefix."""
        genomes = {}
        for prefix in prefixes:
            contigs = FASTA.index_and_collect_contigs(prefixes[prefix]["Reference"])

            genomes[prefix] = {
                "Size": sum(contigs.values()),
                "NContigs": len(contigs),
            }

        return genomes


def _measure_ordering(pair):
    measure = pair[0]
    key = measure.split("(")[0]
    return (_ORDERING[key], measure)


def _get_comment(key):
    comment = _COMMENTS.get(key)
    if comment is None:
        comment = _COMMENTS[key.split("(", 1)[0]]

    return comment


_ORDERING = {
    "lib_type": 00,
    "seq_reads_se": 10,
    "seq_trash_se": 20,
    "seq_trash_se_frac": 30,
    "seq_reads_pairs": 40,
    "seq_trash_pe_1": 50,
    "seq_trash_pe_1_frac": 60,
    "seq_trash_pe_2": 70,
    "seq_trash_pe_2_frac": 80,
    "seq_collapsed": 90,
    "seq_collapsed_frac": 100,
    "seq_retained_reads": 110,
    "seq_retained_nts": 120,
    "seq_retained_length": 130,
    "hits_raw": 140,
    "hits_raw_frac": 150,
    "hits_clonality": 160,
    "hits_unique": 170,
    "hits_unique_frac": 180,
    "hits_coverage": 190,
    "hits_length": 200,
    "ratio_reads": 210,
    "ratio_genome": 220,
}

_COMMENTS = {
    "hits_clonality": "# Fraction of hits that were PCR duplicates",
    "hits_coverage": "# Estimated coverage from unique hits",
    "hits_length": "# Average number of aligned bases per unique hit",
    "hits_raw_frac": "# Total number of hits vs. total number of reads retained",
    "hits_raw": "# Total number of hits (prior to PCR duplicate filtering)",
    "hits_unique_frac": "# Total number of unique hits vs. total number of reads retained",
    "hits_unique": "# Total number of hits (excluding any PCR duplicates)",
    "lib_type": "# SE, PE, or * (for both)",
    "ratio_genome(mito,nuc)": "# Ratio of NTs of unique hits corrected by genome sizes: (NTs(mito) / NTs(nuc)) / (Size(mito) / (2 * Size(nuc)))",
    "ratio_genome(nuc,mito)": "# Ratio of NTs of unique hits corrected by genome sizes: (NTs(nuc) / NTs(mito)) / ((2 * Size(nuc)) / Size(mito))",
    "ratio_reads(nuc,mito)": "# Ratio of unique hits: Hits(nuc) / H(mito)",
    "seq_collapsed_frac": "# Fraction of PE pairs collapsed into one read",
    "seq_collapsed": "# Total number of pairs collapsed into one read",
    "seq_reads_pairs": "# Total number of pairs",
    "seq_reads_se": "# Total number of single-ended reads",
    "seq_retained_length": "# Average number of NTs in retained reads",
    "seq_retained_nts": "# Total number of NTs in retained reads",
    "seq_retained_reads": "# Total number of retained reads",
    "seq_trash_pe_1_frac": "# Fraction of PE mate 1 reads trashed",
    "seq_trash_pe_1": "# Total number of reads",
    "seq_trash_pe_2_frac": "# Fraction of PE mate 2 reads trashed",
    "seq_trash_pe_2": "# Total number of reads",
    "seq_trash_se_frac": "# Fraction of SE reads trashed",
    "seq_trash_se": "# Total number of trashed reads",
}



#!/usr/bin/python3
#
# Copyright (c) 2013 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os
import multiprocessing

import paleomix
import paleomix.common.logging

from paleomix.resources import add_copy_example_command
from paleomix.common.argparse import ArgumentParser, SUPPRESS


_DEFAULT_CONFIG_FILES = [
    "/etc/paleomix/bam_pipeline.ini",
    "~/.paleomix/bam_pipeline.ini",
]


def build_parser(pipeline_variant):
    parser = ArgumentParser(prog="paleomix %s_pipeline" % (pipeline_variant,))

    subparsers = parser.add_subparsers(dest="command", metavar="command")
    add_copy_example_command(subparsers)
    add_makefile_command(subparsers)
    add_run_command(subparsers)

    return parser


def add_makefile_command(subparsers):
    parser = subparsers.add_parser(
        "makefile", help="Print makefile template", aliases=("mkfile",)
    )

    parser.add_argument(
        "samplesheets",
        nargs="*",
        help="Auto-generate targets from illumina samplesheet.csv files",
    )

    parser.add_argument(
        "--minimal",
        default=False,
        action="store_true",
        help="Strip comments from makefile template.",
    )


def add_run_command(subparsers):
    parser = subparsers.add_parser(
        "run",
        aliases=("dryrun",),
        help="Run pipeline on provided makefiles",
        default_config_files=_DEFAULT_CONFIG_FILES,
    )

    parser.add_argument(
        "makefiles",
        nargs="+",
        help="Run pipeline on these makefiles",
        metavar="makefile",
    )

    paleomix.common.logging.add_argument_group(parser)

    group = parser.add_argument_group("Scheduling")
    group.add_argument(
        "--dry-run",
        action="store_true",
        default=False,
        help="If passed, only a dry-run in performed, and no tasks are executed.",
    )
    group.add_argument(
        "--max-threads",
        type=int,
        default=max(2, multiprocessing.cpu_count()),
        help="Max number of threads to use in total",
    )
    group.add_argument(
        "--adapterremoval-max-threads",
        type=int,
        default=min(3, multiprocessing.cpu_count()),
        help="Max number of threads to use per AdapterRemoval instance",
    )
    group.add_argument(
        "--bowtie2-max-threads",
        type=int,
        default=max(1, min(8, multiprocessing.cpu_count() // 2)),
        help="Max number of threads to use per Bowtie2 instance",
    )
    group.add_argument(
        "--bwa-max-threads",
        type=int,
        default=max(1, min(8, multiprocessing.cpu_count() // 2)),
        help="Max number of threads to use per BWA instance",
    )

    group = parser.add_argument_group("Required paths")
    group.add_argument(
        "--jar-root",
        default=os.path.expanduser("~/install/jar_root"),
        help="Folder containing Picard JARs (http://picard.sf.net)",
    )
    group.add_argument(
        "--temp-root",
        default="./temp/",
        type=os.path.abspath,
        help="Location for temporary files and folders",
    )
    group.add_argument(
        "--destination", default=".", help="The destination folder for result files.",
    )

    group = parser.add_argument_group("Files and executables")
    group.add_argument(
        "--list-input-files",
        action="store_true",
        default=False,
        help="List all input files used by pipeline for the "
        "makefile(s), excluding any generated by the "
        "pipeline itself.",
    )
    group.add_argument(
        "--list-output-files",
        action="store_true",
        default=False,
        help="List all output files generated by pipeline for the makefile(s).",
    )
    group.add_argument(
        "--list-executables",
        action="store_true",
        default=False,
        help="List all executables required by the pipeline, "
        "with version requirements (if any).",
    )

    group = parser.add_argument_group("Misc")
    group.add_argument(
        "--jre-option",
        "--jre-options",
        metavar="OPTION",
        dest="jre_options",
        action="append",
        default=[],
        help="May be specified one or more times with options to be passed "
        "to the JRE (Jave Runtime Environment); e.g. to change the "
        "maximum amount of memory (default is -Xmx4g)",
    )
    group.add_argument(
        "--validation",
        metavar="LEVEL",
        type=str.lower,
        choices=("off", "partial", "full"),
        default="full",
        help="Change the amount of validation performed using picard ValidateSamFile. "
        "Set to 'full' to validate both intermediate and final BAM files; set to "
        "'partial' to validate only the final BAM files; and set to 'off' to disable "
        "validation. Reducing the amount of validation may greatly decrease the total "
        "runtime",
    )

    # Removed options
    parser.add_argument("--gatk-max-threads", help=SUPPRESS)
    parser.add_argument("--progress-ui", help=SUPPRESS)
    parser.add_argument("--ui-colors", help=SUPPRESS)



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import sys


_TEMPLATE = """# -*- mode: Yaml; -*-
Project:
  Title: PROJECT_NAME

  # List of samples to be included in the analytical steps, which may be
  # grouped using any arbitrary number of (levels of) groups. (Sub)groups
  # are not required, but may be used instead of listing individual samples
  # in 'ExcludeSamples' and 'FilterSingletons'.
  Samples:
    <GROUP>:
      <SUBGROUP>:
        SAMPLE_NAME:
          # Sex of the sample; used to filter SNPs on homozygous
          # contigs (see below). If not relevant, the value 'NA' may be used.
          Sex:       ...

  # Specifies a set of regions of interest, each representing one or more
  # named regions in a reference sequence (e.g. genes) in BED format.
  RegionsOfInterest:
     NAME:
       # Name of the prefix; is expected to correspond to the filename
       # of the FASTA file without the extension / the name of the
       # prefix used in the BAM pipeline.
       Prefix: PREFIX_NAME
       # Specifies whether or not the sequences are protein coding; if true
       # indels are only included in the final sequence if the length is
       # divisible by 3.
       ProteinCoding: no
       # Include indels in final sequence; note that indels are always called,
       # and used to filter SNPs, even if not included in the resulting FASTA
       # sequences. Requires that 'MultipleSequenceAlignment' is enabled
       IncludeIndels: yes
       # List of contigs for which heterozygous SNPs should be filtered
       # (site set to 'N') based on sex; All sexes used in the 'Samples'
       # section must be listed, except for 'NA' which defaults to no contigs.
       HomozygousContigs:
         Female:
           - chrM
         Male:
           - chrX
           - chrY
           - chrM

  # Filter sites in a sample, replacing any nucleotide not observed
  # in the specified list of samples or groups with 'N'.
#  FilterSingletons:
#    NAME_OF_SAMPLE:
#      - <NAME_OF_GROUP>
#      - NAME_OF_SAMPLE


Genotyping:
  # Default settings for all regions of interest
  Defaults:
    # Regions of interest are expanded by this number of bases when calling
    # SNPs, in order to ensure that adjacent indels can be used during filtering
    # (VCF_filter --min-distance-to-indels and --min-distance-between-indels).
    # The final sequences does not include the padding.
    Padding: 10

    # By default, each set of regions of interest are genotyped seperately,
    # even if these overlap. By setting this option to true, the entire prefix
    # is genotyped once, and all regions of interest are extracted from this.
    # This can only be done for prefixes that only use genotyping defaults.
    GenotypeEntirePrefix: no

    MPileup:
      -E: # extended BAQ for higher sensitivity but lower specificity
      -A: # count anomalous read pairs

    BCFTools:
      -c: # Call genotypes at variant sites using consensus caller (see also -m)

    VCF_Filter:
      # Maximum coverage acceptable for genotyping calls; if set to zero, the
      # default vcf_filter value is used; if set to 'auto', the MaxDepth value
      # will be read from the depth histograms generated by the BAM pipeline.
      MaxReadDepth: 0

      # Minimum coverage acceptable for genotyping calls
      --min-read-depth: 8
      # Min RMS mapping quality
      --min-mapping-quality: 10
      # Min QUAL score (Phred) for genotyping calls
      --min-quality: 30
      # Min distance of variants to indels
      --min-distance-to-indels: 2
      # Min distance between indels
      --min-distance-between-indels: 10
      # Min P-value for strand bias (given PV4)
      --min-strand-bias: 1.0e-4
      # Min P-value for baseQ bias (given PV4)
      --min-baseq-bias: 1.0e-4
      # Min P-value for mapQ bias (given PV4)
      --min-mapq-bias: 1.0e-4
      # Min P-value for end distance bias (given PV4)
      --min-end-distance-bias: 1.0e-4
      # Minimum number of alternative bases observed for variants
      --min-num-alt-bases: 2

# Add / overwrite default settings for a set of regions
#  NAME_OF_REGIONS:
#    ...


MultipleSequenceAlignment:
  # Default settings for all regions of interest
  Defaults:
    Enabled: yes

    # Multiple sequence alignment using MAFFT
    MAFFT:
      # Select alignment algorithm; valid values are 'mafft', 'auto', 'fft-ns-1',
      # 'fft-ns-2', 'fft-ns-i', 'nw-ns-i', 'l-ins-i', 'e-ins-i', and 'g-ins-i'.
      Algorithm: G-INS-i

      # Parameters for mafft algorithm; see above for example of how to specify
      --maxiterate: 1000

# Add / overwrite default settings for a set of regions
#  NAME_OF_REGIONS:
#    ...


PhylogeneticInference:
  PHYLOGENY_NAME:
    # Exclude (groups of) samples from this analytical step
#    ExcludeSamples:
#      - <NAME_OF_GROUP>
#      - NAME_OF_SAMPLE

    # Root the final tree(s) on one or more samples; if no samples
    # are specified, the tree(s) will be rooted on the midpoint(s)
#    RootTreesOn:
#      - <NAME_OF_GROUP>
#      - NAME_OF_SAMPLE

    # If 'yes', a tree is generated per named sequence in the areas of
    # interest; otherwise a super-matrix is created from the combined set
    # of regions specfied below.
    PerGeneTrees: no

    # Which Regions Of Interest to build the phylogeny from.
    RegionsOfInterest:
       REGIONS_NAME:
         # Partitioning scheme for sequences: Numbers specify which group a
         # position belongs to, while 'X' excludes the position from the final
         # partioned sequence; thus "123" splits sequences by codon-positions,
         # while "111" produces a single partition per gene. If set to 'no',
         # a single partition is used for the entire set of regions.
         Partitions: "111"
         # Limit analysis to a subset of a RegionOfInterest; subsets are expected to be
         # located at <genome root>/<prefix>.<region name>.<subset name>.names, and
         # contain single name (corresponding to column 4 in the BED file) per line.
#         SubsetRegions: SUBSET_NAME

    ExaML:
      # Number of times to perform full phylogenetic inference
      Replicates: 1
      # Number of bootstraps to compute
      Bootstraps: 100
      # Model of rate heterogeneity (GAMMA or PSR)
      Model: GAMMA
"""


def main(_argv):
    print(_TEMPLATE)


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



#!/usr/bin/python3
#
# Based on 'FilterUniqueBAM' by
#   Martin Kircher
#   Martin.Kircher@eva.mpg.de
#
"""paleomix rmdup_collapsed [options] < sorted.bam > out.bam

The rmdup_collapsed filters a BAM file for PCR duplicates unpaired reads under
the assumption that any unpaired read have been generated by the merging of
overlapping paired-end reads, and thereby represent the complete template
sequence. PCR duplicates are therefore detected based on both the 5' and 3'
alignment coordinate.

Paired reads (0x1), unmapped reads (0x4), secondary alignments (0x100),
reads that failed QC (0x200), and chimeric alignments (0x800), as identified
using the BAM record flags, are not filtered, but simply written to the output.

By default, filtered reads are flagged using the "duplicate" flag (0x400), and
written to the output. Use the --remove-duplicates command-line option to
instead remove these records from the output.
"""
import collections
import random
import sys

import pysam

from paleomix.common.argparse import ArgumentParser


_FILTERED_FLAGS = 0x1  # PE reads
_FILTERED_FLAGS |= 0x4  # Unmapped
_FILTERED_FLAGS |= 0x100  # Secondary alignment
_FILTERED_FLAGS |= 0x200  # Failed QC
_FILTERED_FLAGS |= 0x800  # Chimeric alignment

_CIGAR_SOFTCLIP = 4
_CIGAR_HARDCLIP = 5


def read_quality(read):
    qualities = read.query_alignment_qualities
    if qualities is None:
        # Generate value in range (-1; 0]
        return -random.random()

    return sum(qualities)


def copy_number(read):
    # has_tag is faster than try/except, since most reads lack the tag.
    if read.has_tag("XP"):
        return read.get_tag("XP")

    return 0


def mark_duplicate_reads(reads):
    """Identifies the best read from a set of PCR duplicates, and marks all
    other reads as duplicates. The best read is selected by quality, among the
    reads sharing the most common CIGAR string.
    """
    by_cigar = collections.defaultdict(list)
    for read in reads:
        key = tuple(read.cigartuples)
        by_cigar[key].append(read)

    # Select the most common CIGAR strings, favoring simple CIGARs
    best_count, best_cigar_len = max(
        (len(values), -len(cigar)) for cigar, values in by_cigar.items()
    )
    best_cigar_len = -best_cigar_len

    best_read = None
    best_quality = -1
    copies = len(reads)

    for cigar, candidates in by_cigar.items():
        if len(cigar) == best_cigar_len and len(candidates) == best_count:
            for read in candidates:
                copies += copy_number(read)
                quality = read_quality(read)

                if quality > best_quality:
                    best_read = read
                    best_quality = quality
        else:
            copies += sum(copy_number(read) for read in candidates)

    best_read.set_tag("XP", copies, "i")
    for read in reads:
        read.is_duplicate = read is not best_read


def write_read(args, out, read_and_alignment, duplicates_by_alignment):
    read, alignment = read_and_alignment
    if alignment is not None:
        duplicates = duplicates_by_alignment.pop(alignment)

        if len(duplicates) > 1:
            # Select the best read and mark the others as duplicates.
            mark_duplicate_reads(duplicates)
        else:
            duplicates[0].is_duplicate = False
            duplicates[0].set_tag("XP", 1, "i")

    if not (args.remove_duplicates and read.is_duplicate):
        out.write(read)


def can_write_read(read_and_alignment, current_position):
    """Returns true if the first read in the cache can safely be written. This
    will be the case if the read was not the first in a set of reads with the
    same alignment, or if the current position has gone beyond the last base
    covered in that alignment.
    """
    _, alignment = read_and_alignment
    if alignment is None:
        return True

    current_ref_id, current_ref_start = current_position
    alignment_ref_id, _, _, alignment_ref_end = alignment

    return alignment_ref_id != current_ref_id or alignment_ref_end < current_ref_start


def clipped_bases_at_front(cigartuples):
    """Returns number of bases soft or hard clipped at start of the CIGAR."""
    total = 0
    for (operation, length) in cigartuples:
        if operation != _CIGAR_SOFTCLIP and operation != _CIGAR_HARDCLIP:
            break

        total += length

    return total


def unclipped_alignment_coordinates(read):
    """Returns tuple describing the alignment, with external coordinates
    modified to account for clipped bases, assuming an ungapped alignment to
    the reference. This is equivalent to the behavior of Picard MarkDuplicates.
    """
    cigartuples = read.cigartuples
    start = read.reference_start - clipped_bases_at_front(cigartuples)
    end = read.reference_end + clipped_bases_at_front(reversed(cigartuples))

    return (read.reference_id, read.is_reverse, start, end)


def process_aligned_read(cache, duplicates_by_alignment, read):
    """Processes an aligned read, either pairing it with an existing read, or
    creating a new alignment block to track copies of this copies.
    """
    alignment = unclipped_alignment_coordinates(read)

    try:
        duplicates_by_alignment[alignment].append(read)
        cache.append((read, None))
    except KeyError:
        # No previous reads with matching alignment; this read will
        # serve to track any other reads with the same alignment.
        duplicates_by_alignment[alignment] = [read]
        cache.append((read, alignment))


def is_trailing_unmapped_read(read):
    return read.is_unmapped and read.reference_id == -1 and read.reference_start == -1


def process(args, infile, outfile):
    cache = collections.deque()
    duplicates_by_alignment = {}
    last_position = (0, 0)
    read_num = 1

    for read_num, read in enumerate(infile, start=read_num):
        current_position = (read.reference_id, read.reference_start)
        if last_position > current_position:
            # Check also catches trailing unmapped reads mapped to (-1, -1).
            if not is_trailing_unmapped_read(read):
                sys.stderr.write(
                    "ERROR: Input file is not sorted by "
                    "coordinates at read %i. Aborting!\n" % (read_num,)
                )
                return 1

            cache.append((read, None))
            break
        elif read.flag & _FILTERED_FLAGS:
            cache.append((read, None))
        else:
            process_aligned_read(cache, duplicates_by_alignment, read)

        last_position = current_position
        while cache and can_write_read(cache[0], current_position):
            write_read(args, outfile, cache.popleft(), duplicates_by_alignment)

    while cache:
        write_read(args, outfile, cache.popleft(), duplicates_by_alignment)

    assert not duplicates_by_alignment, duplicates_by_alignment

    for read_num, read in enumerate(infile, start=read_num + 1):
        if not is_trailing_unmapped_read(read):
            sys.stderr.write(
                "ERROR: Input file is not sorted by "
                "coordinates at read %i. Aborting!\n" % (read_num,)
            )
            return 1

        outfile.write(read)

    return 0


def parse_args(argv):
    parser = ArgumentParser(prog="paleomix rmdup_collapsed", usage=__doc__)
    parser.add_argument(
        "input",
        default="-",
        nargs="?",
        help="BAM file; if not set, input is read from STDIN.",
    )
    parser.add_argument(
        "--remove-duplicates",
        help="Remove duplicates from output; by default "
        "duplicates are only flagged (flag = 0x400).",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "--seed",
        default=None,
        type=int,
        help="Seed used for randomly selecting representative reads, when reads do not "
        "have quality scores assigned. Is initialized using system time by default.",
    )

    return parser.parse_args(argv)


def main(argv):
    args = parse_args(argv)

    # Initialize seed used when selecting among reads without quality scores
    random.seed(args.seed)

    if args.input == "-" and sys.stdin.isatty():
        sys.stderr.write("STDIN is a terminal, terminating!\n")
        return 1
    elif sys.stdout.isatty():
        sys.stderr.write("STDOUT is a terminal, terminating!\n")
        return 1

    with pysam.AlignmentFile(args.input, "rb") as infile:
        with pysam.AlignmentFile("-", "wb", template=infile) as outfile:
            return process(args, infile, outfile)

    return 0


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os
import random
import collections

from paleomix.nodes.formats import FastaToPartitionedInterleavedPhyNode as ToPhylipNode
from paleomix.nodes.raxml import RAxMLParsimonyTreeNode
from paleomix.nodes.phylip import PHYLIPBootstrapNode
from paleomix.nodes.examl import ExaMLNode, ExaMLParserNode
from paleomix.nodes.newick import NewickRerootNode, NewickSupportNode
from paleomix.common.fileutils import swap_ext, add_postfix


def _build_supermatrix(
    destination, input_files, exclude_samples, subset_files, dependencies
):
    matrixprefix = os.path.join(destination, "alignments")

    # This supermatrix (and partitions) is not used by the pipeline, but is
    # built for reference purposes; may be made optional in the future.
    supermatrix = ToPhylipNode(
        infiles=input_files,
        out_prefix=matrixprefix,
        exclude_groups=exclude_samples,
        dependencies=dependencies,
        file_dependencies=subset_files,
    )

    # Supermatrix with empty columns (all N, n, or -) stripped
    reduced = ToPhylipNode(
        reduce=True,
        infiles=input_files,
        out_prefix=matrixprefix + ".reduced",
        exclude_groups=exclude_samples,
        dependencies=dependencies,
        file_dependencies=subset_files,
    )

    return (supermatrix, reduced)


def _examl_nodes(
    options,
    settings,
    input_alignment,
    input_partitions,
    input_binary,
    output_template,
    dependencies,
):
    parsimony_tree = output_template % ("parsimony_tree",)
    tree = RAxMLParsimonyTreeNode(
        input_alignment=input_alignment,
        input_partitions=input_partitions,
        output_tree=parsimony_tree,
        dependencies=dependencies,
    )

    return ExaMLNode(
        input_binary=input_binary,
        initial_tree=parsimony_tree,
        output_template=output_template,
        model=settings["ExaML"]["Model"].upper(),
        threads=options.examl_max_threads,
        dependencies=tree,
    )


def _build_rerooted_trees(nodes, reroot_on):
    filenames = []
    for node in nodes:
        for filename in node.output_files:
            if filename.endswith(".result"):
                filenames.append(filename)

    output_file = os.path.dirname(filenames[0]) + ".newick"

    return NewickRerootNode(
        tree_files=filenames,
        output_file=output_file,
        taxa=reroot_on,
        dependencies=nodes,
    )


def _build_examl_replicates(
    options, phylo, destination, input_alignment, input_partition, dependencies
):
    input_binary = os.path.join(destination, "alignments.reduced.binary")
    binary = ExaMLParserNode(
        input_alignment=input_alignment,
        input_partition=input_partition,
        output_file=input_binary,
        dependencies=dependencies,
    )

    replicates = []
    for replicate_num in range(phylo["ExaML"]["Replicates"]):
        replicate_destination = os.path.join(destination, "replicates")
        replicate_template = os.path.join(
            replicate_destination, "replicate.%04i.%%s" % (replicate_num,)
        )
        replicates.append(
            _examl_nodes(
                options,
                phylo,
                input_alignment,
                input_partition,
                input_binary,
                replicate_template,
                binary,
            )
        )

    if replicates:
        return _build_rerooted_trees(replicates, phylo["RootTreesOn"])

    return None


def _build_examl_bootstraps(
    options, phylo, destination, input_alignment, input_partition, dependencies
):
    bootstraps = []
    num_bootstraps = phylo["ExaML"]["Bootstraps"]
    bootstrap_destination = os.path.join(destination, "bootstraps")
    bootstrap_template = os.path.join(bootstrap_destination, "bootstrap.%04i.phy")

    for bootstrap_num in range(num_bootstraps):
        bootstrap_alignment = bootstrap_template % (bootstrap_num,)
        bootstrap = PHYLIPBootstrapNode(
            input_alignment=input_alignment,
            input_partition=input_partition,
            output_alignment=bootstrap_alignment,
            seed=random.randint(1, 2 ** 32 - 1),
            dependencies=dependencies,
        )

        bootstrap_binary = swap_ext(bootstrap_alignment, ".binary")
        bootstrap_final = swap_ext(bootstrap_alignment, ".%s")
        bs_binary = ExaMLParserNode(
            input_alignment=bootstrap_alignment,
            input_partition=input_partition,
            output_file=bootstrap_binary,
            dependencies=bootstrap,
        )

        bootstraps.append(
            _examl_nodes(
                options=options,
                settings=phylo,
                input_alignment=bootstrap_alignment,
                input_partitions=input_partition,
                input_binary=bootstrap_binary,
                output_template=bootstrap_final,
                dependencies=bs_binary,
            )
        )

    if bootstraps:
        return _build_rerooted_trees(bootstraps, phylo["RootTreesOn"])

    return None


def add_bootstrap_support(destination, replicate, bootstrap):
    if not (replicate and bootstrap):
        return [_f for _f in (replicate, bootstrap) if _f]

    replicate_file = os.path.join(destination, "replicates.newick")
    bootstrap_file = os.path.join(destination, "bootstraps.newick")
    output_file = add_postfix(replicate_file, ".support")

    return (
        NewickSupportNode(
            main_tree_files=replicate_file,
            support_tree_files=bootstrap_file,
            output_file=output_file,
            dependencies=(bootstrap, replicate),
        ),
    )


def _build_examl_nodes(
    options, settings, destination, input_files, subset_files, dependencies
):
    input_alignment = os.path.join(destination, "alignments.reduced.phy")
    input_partition = os.path.join(destination, "alignments.reduced.partitions")

    excluded = settings["ExcludeSamples"]
    supermatrix = _build_supermatrix(
        destination, input_files, excluded, subset_files, dependencies
    )
    examl_args = (
        options,
        settings,
        destination,
        input_alignment,
        input_partition,
        supermatrix,
    )

    examl_replicates = _build_examl_replicates(*examl_args)
    examl_bootstraps = _build_examl_bootstraps(*examl_args)
    examl_dependencies = add_bootstrap_support(
        destination, examl_replicates, examl_bootstraps
    )

    return examl_dependencies


def _build_examl_per_gene_nodes(
    options, settings, run_dd, roi, destination, filtering, dependencies
):
    regions = settings["Project"]["Regions"][roi["Name"]]
    sequences = regions["Sequences"][roi["SubsetRegions"]]
    subset_files = regions["SubsetFiles"][roi["SubsetRegions"]]
    filtering_postfix = ".filtered" if any(filtering.values()) else ""
    sequence_dir = os.path.join(
        options.destination, "alignments", roi["Name"] + filtering_postfix
    )
    msa_enabled = settings["MultipleSequenceAlignment"][regions["Name"]]["Enabled"]
    fasta_extension = ".afa" if msa_enabled else ".fasta"

    partitions = roi["Partitions"] or "111"

    nodes = []
    for sequence in sequences:
        seq_source = os.path.join(sequence_dir, sequence + fasta_extension)
        seq_destination = os.path.join(destination, sequence)
        input_files = {sequence: {"partitions": partitions, "filenames": [seq_source]}}
        nodes.extend(
            _build_examl_nodes(
                options,
                run_dd,
                seq_destination,
                input_files,
                subset_files,
                dependencies,
            )
        )

    return nodes


def _build_examl_regions_nodes(
    options, settings, run_dd, destination, filtering, dependencies
):
    input_files = collections.defaultdict(dict)
    subset_files = []
    for (roi_name, roi_dd) in run_dd["RegionsOfInterest"].items():
        regions = settings["Project"]["Regions"][roi_name]
        subset_key = roi_dd.get("SubsetRegions")
        sequences = regions["Sequences"][subset_key]
        subset_files.extend(regions["SubsetFiles"][subset_key])

        partitions = roi_dd["Partitions"]
        filtering_postfix = ".filtered" if any(filtering.values()) else ""
        sequence_dir = os.path.join(
            options.destination, "alignments", roi_name + filtering_postfix
        )
        msa_enabled = settings["MultipleSequenceAlignment"][regions["Name"]]["Enabled"]
        fasta_extension = ".afa" if msa_enabled else ".fasta"

        if partitions:
            for sequence in sequences:
                seq_source = os.path.join(sequence_dir, sequence + fasta_extension)

                input_files[sequence] = {
                    "partitions": partitions,
                    "filenames": [seq_source],
                }
        else:
            filenames = []
            for sequence in sequences:
                filenames.append(os.path.join(sequence_dir, sequence + fasta_extension))

            input_files[roi_name] = {"partitions": "1", "filenames": filenames}

    return _build_examl_nodes(
        options, run_dd, destination, dict(input_files), subset_files, dependencies
    )


def build_phylogeny_nodes(options, settings, filtering, dependencies):
    nodes = []
    for (run_name, run_dd) in settings["PhylogeneticInference"].items():
        destination = os.path.join(options.destination, "phylogenies", run_name)

        if run_dd["PerGeneTrees"]:
            run_nodes = []
            for roi in run_dd["RegionsOfInterest"].values():
                roi_destination = os.path.join(destination, roi["Name"])
                run_nodes.extend(
                    _build_examl_per_gene_nodes(
                        options,
                        settings,
                        run_dd,
                        roi,
                        roi_destination,
                        filtering,
                        dependencies,
                    )
                )
            nodes.extend(run_nodes)
        else:
            nodes.extend(
                _build_examl_regions_nodes(
                    options, settings, run_dd, destination, filtering, dependencies
                )
            )

    return nodes


def chain_examl(_pipeline, options, makefiles):
    destination = options.destination  # Move to makefile
    for makefile in makefiles:
        filtering = makefile["Project"]["FilterSingletons"]
        options.destination = os.path.join(destination, makefile["Project"]["Title"])

        makefile["Nodes"] = build_phylogeny_nodes(
            options, makefile, filtering, makefile["Nodes"]
        )
    options.destination = destination



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import sys
import collections

from paleomix.common.utilities import get_in, set_in
from paleomix.common.text import parse_padded_table

from paleomix.tools.bam_stats.common import BAMStatsError


##############################################################################
##############################################################################
##


class ReadGroup:
    __slots__ = ["SE", "PE_1", "PE_2", "Collapsed", "Hits", "M", "I", "D", "Size"]

    def __init__(self):
        self.SE = 0
        self.PE_1 = 0
        self.PE_2 = 0
        self.Collapsed = 0
        self.Hits = 0
        self.M = 0
        self.I = 0
        self.D = 0
        self.Size = 0

    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def add(self, other):
        self.SE += other.SE
        self.PE_1 += other.PE_1
        self.PE_2 += other.PE_2
        self.Collapsed += other.Collapsed
        self.Hits += other.Hits
        self.M += other.M
        self.I += other.I
        self.D += other.D
        self.Size += other.Size

    def has_values(self):
        for slot in ReadGroup.__slots__:
            if slot != "Size" and getattr(self, slot):
                return True

        return False


# Header prepended to output tables
TABLE_HEADER = """# Columns:
#   Contig:    Contig, chromosome, or feature for which a depth histogram was
#              created. Unnamed features are named after the chromosome or
#              contig on which they are located, with a star appended. For
#              example "chr1*".
#   Size:      The total size of the region. Multiple features with the same
#              name are combined into one row, with the size representing to
#              total of these. Note that overlapping bases are counted 2 (or
#              more) times.
#   Hits:      Sum of SE, PE_1, and PE_2 hits. Note that supplementary
#              alignments, duplicates, reads that failed QC, secondary
#              alignments, and unmapped reads are ignored.
#   SE, PE_*:  Number of Single Ended, and Pair Ended (mate 1 and 2) hits
#              overlapping the current contig or intervals. Note that a hit
#              may be counted multiple times if it overlaps multiple intervals
#   Collapsed: Number of hits for PE pair collapsed into a single read.
#   M, I, D:   Number of aligned (M), inserted (I) and deleted (D) bases
#              relative to references.
#   Coverage:  Average number of bases covering each position in the
#              contig(s)/intervals(s).
"""


def calculate_totals(table):
    lengths = {}
    for samples in table.values():
        for libraries in samples.values():
            for contigs in libraries.values():
                for (name, contig) in contigs.items():
                    size = lengths.get(name)
                    if (size is not None) and (size != contig.Size):
                        raise BAMStatsError(name)
                    lengths[name] = contig.Size

    for (name, samples) in sorted(table.items()):
        for (sample, libraries) in sorted(samples.items()):
            for (library, contigs) in sorted(libraries.items()):
                totals = _calculate_totals_in(contigs, lengths)
                set_in(table, (name, sample, library), totals)

            totals = _calculate_totals_in(libraries, lengths)
            set_in(table, (name, sample, "*"), totals)

        set_in(table, (name, "*", "*"), _calculate_totals_in(table, lengths))
    return table


def build_rows(table):
    yield (
        "Name",
        "Sample",
        "Library",
        "Contig",
        "Size",
        "Hits",
        "SE",
        "PE_1",
        "PE_2",
        "Collapsed",
        "M",
        "I",
        "D",
        "Coverage",
    )

    for (name, samples) in sorted(table.items()):
        for (sample, libraries) in sorted(samples.items()):
            for (library, contigs) in sorted(libraries.items()):
                for (contig, subtable) in sorted(contigs.items()):
                    row = [
                        name,
                        sample,
                        library,
                        contig,
                        subtable.Size,
                        subtable.SE
                        + subtable.PE_1
                        + subtable.PE_2
                        + subtable.Collapsed,
                        subtable.SE,
                        subtable.PE_1,
                        subtable.PE_2,
                        subtable.Collapsed,
                        subtable.M,
                        subtable.I,
                        subtable.D,
                        float(subtable.M) / subtable.Size,
                    ]
                    yield row
                yield "#"
            yield "#"


def read_table(table, filename):
    with open(filename) as table_file:
        for record in parse_padded_table(table_file):
            key = (
                record["Name"],
                record["Sample"],
                record["Library"],
                record["Contig"],
            )
            if "*" in key:
                continue

            subtable = get_in(table, key)
            if subtable is None:
                subtable = ReadGroup()
                subtable.Size = int(record["Size"])
                set_in(table, key, subtable)

            assert int(subtable.Size) == int(record["Size"])
            for key in ReadGroup.__slots__:
                if key != "Size":
                    subtable[key] += int(record.get(key, 0))


def write_table(table, filename):
    table = calculate_totals(table)
    rows = build_rows(table)

    if filename == "-":
        output_handle = sys.stdout
    else:
        output_handle = open(filename, "w")

    try:
        output_handle.write(TABLE_HEADER)
        for line in rows:
            output_handle.write("\t".join(map(str, line)))
            output_handle.write("\n")
    finally:
        if output_handle is not sys.stdout:
            output_handle.close()


def _calculate_totals_in(tables, lengths):
    totals = collections.defaultdict(ReadGroup)
    total_size = sum(lengths.values())

    subtables = list(tables.items())
    while subtables:
        subtable_key, subtable = subtables.pop()
        if subtable_key == "*":
            totals[subtable_key].Size = total_size
        elif isinstance(subtable, ReadGroup):
            totals[subtable_key].add(subtable)
            totals["*"].add(subtable)

            totals[subtable_key].Size = lengths[subtable_key]
            totals["*"].Size = total_size
        else:
            subtables.extend(subtable.items())

    return dict(totals)



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Wrapper and utility functions for VCF handling, using
the VCF data-structures from pysam."""

import collections


Indel = collections.namedtuple(
    "Indel", ["in_reference", "pos", "prefix", "what", "postfix"]
)


def parse_indel(vcf):
    """Parses the VCF record of an indel, and returns a tuple containing the
    position (0-based) of the previous base, a boolean indicating whether or
    not the subsequent sequence is found in the reference sequence, and a
    string containing the bases added to / removed from the reference.

    Thus (7, False, "ACGT", "AC") indicates that the sequence ACGT has been
    inserted following the 8th nucleotide, compared with the reference, and
    that the insertion is followed by the bases "AC" on the reference."""
    if not is_indel(vcf):
        raise ValueError("SNP passed to 'parse_indel'!")
    elif "," in vcf.alt:
        raise ValueError("VCF records with multiple indels not supported!")
    elif vcf.ref[0] != vcf.alt[0]:
        raise ValueError(
            "Sequences do not match VCF spec, first base differs: "
            "%s:%s -- %s > %s" % (vcf.contig, vcf.pos + 1, vcf.ref, vcf.alt)
        )

    ref_len, alt_len = len(vcf.ref), len(vcf.alt)
    # The length of the insertion / deletion
    len_diff = abs(alt_len - ref_len)

    # Wheter or not the sequence 'what' is found in the reference
    in_reference = ref_len >= alt_len

    # The sequence added or removed from the reference
    longest = max(vcf.ref, vcf.alt, key=len)
    shortest = min(vcf.ref, vcf.alt, key=len)
    what = longest[1 : len_diff + 1]

    postfix = shortest[1:]
    if longest[len_diff + 1 :] != postfix:
        raise ValueError("Sequence postfix does not match; malformed indel!")

    return Indel(in_reference, vcf.pos, vcf.ref[0], what, postfix)


def is_indel(vcf):
    """Returns true if the VCF entry represents an indel."""
    # FIXME: Is this a universal key for indels?
    return "INDEL" in vcf.info


# The corresponding nucleotides for each value in the VCF PL field
_genotype_indices = [(jj, ii) for ii in range(0, 10) for jj in range(0, ii + 1)]


def get_ml_genotype(vcf, sample=0):
    """Returns the most likely genotype of a sample in a vcf record. If no
    single most likely genotype can be determined, the function returns 'N' for
    both bases."""
    genotypes = []
    genotypes.extend(vcf.ref.split(","))
    genotypes.extend(vcf.alt.split(","))

    PL = list(map(int, get_format(vcf, sample)["PL"].split(",")))

    if len(PL) == len(genotypes):
        ploidy = 1
    else:
        expected_length = (len(genotypes) * (len(genotypes) + 1)) // 2
        if len(PL) != expected_length:
            raise ValueError(
                "Expected %i PL values, found %i" % (expected_length, len(PL))
            )
        ploidy = 2

    if PL.count(min(PL)) > 1:
        # No single most likely genotype
        return ("N", "N")

    most_likely = min(range(len(PL)), key=PL.__getitem__)
    if ploidy == 1:
        prefix = postfix = most_likely
    else:
        prefix, postfix = _genotype_indices[most_likely]

    return (genotypes[prefix], genotypes[postfix])


def get_format(vcf, sample=0):
    return dict(zip(vcf.format.split(":"), vcf[sample].split(":")))



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""
Reads SAM or BAM from STDIN and outputs cleaned and sorted BAM to STDOUT.

The cleanup involves setting header-flags, fixing mate information, clearing
various fields of unmapped reads (due to BWA leaving CIGARs in unmapped reads),
sorting the input, and updating MD and NM tags.

This script also solves a problem with parsing record-less files by SAMTools,
which fails with a parse error if the input SAM file only contains a header.
This is mainly a problem when aligning ancient DNA lanes against mt genomes,
as there may not be any hits in an entire lane. For example, the following
command will not work:
$ samtools view -H INPUT.BAM | samtools view -Sbu -

"""
import sys

import pysam

import paleomix.tools.factory

import paleomix.common.argparse as argparse
import paleomix.common.procs as processes


# Mask to select flags that are relevant to SE reads; this excludes flags where
# no assumptions can if 0x1 is not set, per the SAM specification (see below).
_SE_FLAGS_MASK = ~(0x2 | 0x8 | 0x20 | 0x40 | 0x80)


def _set_sort_order(header):
    """Updates a BAM header to indicate coordinate sorting."""
    hd_dict = header.setdefault("HD", {"GO": "none", "VN": "1.0"})
    hd_dict["SO"] = "coordinate"


def _set_pg_tags(header, tags):
    """Updates PG tags in a BAM header, taking a sequence of ID:TAG:VALUEs."""
    for tag in tags:
        pg_id, pg_field, pg_value = tag.split(":")

        for pg_dict in header.setdefault("PG", []):
            if pg_dict.get("ID") == pg_id:
                pg_dict[pg_field] = pg_value
                break
        else:
            header["PG"].append({"ID": pg_id, pg_field: pg_value})


def _set_rg_tags(header, rg_id, rg_tags):
    """Updates RG tags in a BAM header, taking a sequence of TAG:VALUEs."""
    readgroup = {"ID": rg_id}
    for tag in rg_tags:
        rg_field, rg_value = tag.split(":")
        readgroup[rg_field] = rg_value
    header["RG"] = [readgroup]


def _cleanup_record(record):
    """Cleans up the properties of a BAM record, ensuring that only appropriate
    flags and values are set, such that the record follows section 1.4 of the
    SAM specification (https://samtools.github.io/hts-specs/SAMv1.pdf). The
    record itself (or a new record) is returned.
    """
    if not record.is_paired:
        # Unset 0x2 (properly aligned), 0x8 (next mate unmapped),
        # 0x20 (next mate reverse), 0x40 (first mate), and 0x80 (last mate).
        record.flag = record.flag & (~0xEA)
        record.rnext = -1
        record.pnext = -1
        record.tlen = 0
    elif record.mate_is_unmapped and record.has_tag("MC"):
        # Picard ValidateSamFile (2.9.1) objects to MC tags for unmapped mates,
        # which are currently added by SAMTools (v1.4).
        tags = record.get_tags(with_value_type=True)
        record.set_tags([tag for tag in tags if tag[0] != "MC"])

    if record.is_unmapped:
        record.mapq = 0
        record.cigar = None
        # Unset 0x2 (properly aligned), 0x100 (secondary), and 0x800 (chimeric)
        record.flag = record.flag & (~0x902)

        if record.mate_is_unmapped:
            record.rnext = -1
            record.pnext = -1

        # Per the spec, unmapped reads should be placed with their mate
        record.tid = record.rnext
        record.pos = record.pnext
        record.tlen = 0
    elif record.mate_is_unmapped:
        record.rnext = record.tid
        record.pnext = record.pos
        record.tlen = 0

    return record


def _filter_record(args, record):
    """Returns True if the record should be filtered (excluded), based on the
    --exclude-flags and --require-flags options. Certain flags are ignored when
    filtering SE reads, namely those not included in _SE_FLAGS_MASK (above).
    """
    if record.is_paired:
        exclude_flags = args.exclude_flags
        require_flags = args.require_flags
    else:
        exclude_flags = args.exclude_flags & _SE_FLAGS_MASK
        require_flags = args.require_flags & _SE_FLAGS_MASK

    if record.flag & exclude_flags:
        return True
    elif ~(record.flag & require_flags) & require_flags:
        return True

    return False


def _cleanup_unmapped(args):
    """Reads a BAM (or SAM, if cleanup_sam is True) file from STDIN, and
    filters reads according to the filters specified in the commandline
    arguments 'args'. The resulting records are written to STDOUT in
    uncompressed BAM format. The output BAM is marked as sorted (under the
    assumption that 'samtools sort' is to be run on the output) and PG tags are
    updated if specified in the args.
    """

    filter_by_flag = bool(args.exclude_flags or args.require_flags)
    with pysam.AlignmentFile("-") as input_handle:
        header = dict(input_handle.header)
        _set_sort_order(header)
        _set_pg_tags(header, args.update_pg_tag)
        if args.rg_id is not None:
            _set_rg_tags(header, args.rg_id, args.rg)

        with pysam.AlignmentFile("-", "wbu", header=header) as output_handle:
            for record in input_handle:
                # Ensure that the properties make sense before filtering
                record = _cleanup_record(record)

                if not record.is_unmapped and (record.mapq < args.min_quality):
                    continue
                elif filter_by_flag and _filter_record(args, record):
                    continue

                if args.rg_id is not None:
                    # Ensure that only one RG tag is set
                    tags = record.get_tags(with_value_type=True)
                    tags = [tag for tag in tags if tag[0] != "RG"]
                    tags.append(("RG", args.rg_id, "Z"))
                    record.set_tags(tags)

                output_handle.write(record)

    return 0


def _build_wrapper_command(args):
    bam_cleanup = paleomix.tools.factory.new("cleanup")
    if args.fasta is not None:
        bam_cleanup.set_option("--fasta", args.fasta)
    bam_cleanup.set_option("--temp-prefix", args.temp_prefix)
    bam_cleanup.set_option("--min-quality", str(args.min_quality))
    bam_cleanup.set_option("--exclude-flags", hex(args.exclude_flags))

    for value in args.update_pg_tag:
        bam_cleanup.add_option("--update-pg-tag", value)

    if args.rg_id is not None:
        bam_cleanup.set_option("--rg-id", args.rg_id)
        for value in args.rg:
            bam_cleanup.add_option("--rg", value)

    return bam_cleanup.call


def _run_cleanup_pipeline(args):
    bam_cleanup = _build_wrapper_command(args)
    commands = []

    try:
        if args.paired_end:
            # Convert input to (uncompressed) BAM and fix mate information for PE reads
            commands.append(["samtools", "fixmate", "-O", "bam", "-", "-"])

        # Cleanup / filter reads. Must be done after 'fixmate', as BWA may produce
        # hits where the mate-unmapped flag is incorrect, which 'fixmate' fixes.
        commands.append(bam_cleanup + ["cleanup"])

        # Sort by coordinates and output uncompressed BAM
        commands.append(
            ["samtools", "sort", "-l", "0", "-O", "bam", "-T", args.temp_prefix]
        )

        # Update NM and MD tags; output BAM (-b) to stdout
        if args.fasta is not None:
            commands.append(["samtools", "calmd", "-b", "-", args.fasta])

        procs = []
        last_out = sys.stdin
        for cmd in commands:
            proc_stdout = None if cmd is commands[-1] else processes.PIPE
            procs.append(processes.open_proc(cmd, stdin=last_out, stdout=proc_stdout))

            last_out.close()
            last_out = procs[-1].stdout

        return int(any(processes.join_procs(procs)))
    except Exception:
        for proc in procs.values():
            proc.terminate()
        raise


def parse_args(argv):
    """Parses a list of command-line arguments, excluding the program name."""
    prog = "paleomix cleanup"
    usage = "%s --temp-prefix prefix --fasta reference.fasta < in.sam" % (prog,)

    parser = argparse.ArgumentParser(prog=prog, usage=usage)
    # "Hidden" commands, invoking the various sub-parts of this script
    parser.add_argument("command", nargs="?", help=argparse.SUPPRESS)

    parser.add_argument(
        "--fasta",
        default=None,
        help="Reference FASTA sequence; if set, the calmd command is used to "
        "re-calculate MD tags.",
    )
    parser.add_argument(
        "--temp-prefix", required=True, help="REQUIRED: Prefix for temp files"
    )
    parser.add_argument(
        "-q",
        "--min-quality",
        type=int,
        default=0,
        help="Exclude aligned reads with a mapping quality below this value; note "
        "that this filter ONLY applies to aligned reads",
    )
    parser.add_argument(
        "-f",
        "--require-flags",
        default=0,
        type=lambda value: int(value, 0),  # Handle hex, etc.
        help="Only include reads with all of these flags set; note that flags only "
        "valid for paired-end reads (0x2, 0x8, 0x20, 0x40, 0x80) are ignored when "
        "processing single-end reads.",
    )
    parser.add_argument(
        "-F",
        "--exclude-flags",
        default=0,
        type=lambda value: int(value, 0),  # Handle hex, etc.
        help="Exclude reads with any of these flags set; note that flags only valid "
        "for paired-end reads (0x2, 0x8, 0x20, 0x40, 0x80) are ignored when "
        "processing single-end reads.",
    )
    parser.add_argument(
        "--paired-end",
        default=False,
        action="store_true",
        help="If enabled, additional processing of PE reads is carried out, including "
        "updating of mate information [Default: off]",
    )

    parser.add_argument(
        "--update-pg-tag",
        default=[],
        action="append",
        help="Update one PG tags with the given values, creating the tag if it does "
        'not already exist. Takes arguments in the form "PGID:TAG:VALUE".',
    )
    parser.add_argument(
        "--rg-id",
        default=None,
        help="If set, the read-group is overwritten based on tags set using the --rg "
        "option, using the id specified using --rg-id.",
    )
    parser.add_argument(
        "--rg",
        default=[],
        action="append",
        help="Readgroup in the form 'ID:TAG:VALUE'.",
    )

    args = parser.parse_args(argv)
    if args.command not in (None, "cleanup"):
        parser.error("unrecognized arguments: %s" % (args.command,))

    return args


def main(argv):
    """Main function; returns 0 on success, non-zero otherwise."""
    args = parse_args(argv)

    if args.command == "cleanup":
        return _cleanup_unmapped(args)
    elif args.command:
        raise NotImplementedError("Unexpected command %r" % (args.command,))

    sys.stderr.write("Reading SAM file from STDIN\n")
    return _run_cleanup_pipeline(args)



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os

from paleomix.nodes.adapterremoval import SE_AdapterRemovalNode, PE_AdapterRemovalNode
from paleomix.nodes.validation import ValidateFASTQFilesNode


class Reads:
    def __init__(self, config, record, quality_offset):
        self.quality_offset = quality_offset
        self.files = {}
        self.stats = None
        self.validation = ()
        self.nodes = ()

        tags = record["Tags"]
        self.folder = os.path.join(
            config.destination,
            tags["Target"],
            "reads",
            tags["SM"],
            tags["LB"],
            tags["PU"],
            tags["Folder"],
        )

        lane_type = record.get("Type")
        if lane_type == "Raw":
            self._init_raw_reads(config, record)
        elif lane_type == "Trimmed":
            self._init_pretrimmed_reads(record)
        else:
            assert False, "Unexpected data type in Reads(): %s" % (repr(lane_type))

        for name, value in record["Options"]["ExcludeReads"].items():
            if value:
                self.files.pop(name, None)

    def _init_pretrimmed_reads(self, record):
        self.files.update(record["Data"])

        nodes = []
        validation = []
        for key, templ in self.files.items():
            if key == "Paired":
                input_files = [
                    templ.format(Pair=1),
                    templ.format(Pair=2),
                ]
            else:
                input_files = [templ]

            output_file = os.path.join(self.folder, key.lower() + ".statistics")
            node = ValidateFASTQFilesNode(
                input_files=input_files,
                output_file=output_file,
                offset=self.quality_offset,
                collapsed=("Collapsed" in key),
            )

            nodes.append(node)
            validation.append(output_file)

        self.nodes = tuple(nodes)
        self.validation = tuple(validation)

    def _init_raw_reads(self, config, record):
        ar_options = dict(record["Options"]["AdapterRemoval"])
        # Setup of "--collapsed" is handled by the node itself
        collapse_reads = ar_options.pop("--collapse")
        collapse_reads = collapse_reads or collapse_reads is None

        output_quality = self.quality_offset
        if output_quality == "Solexa":
            output_quality = "64"

        ar_options["--qualitybase"] = self.quality_offset
        ar_options["--qualitybase-output"] = output_quality

        init_args = {
            "output_prefix": os.path.join(self.folder, "reads"),
            "threads": config.adapterremoval_max_threads,
            "options": ar_options,
        }

        output_tmpl = "{output_prefix}.%s.gz".format(**init_args)

        if "SE" in record["Data"]:
            self.files["Single"] = output_tmpl % ("truncated",)
            init_args["input_file"] = record["Data"]["SE"]
            command = SE_AdapterRemovalNode(**init_args)
        else:
            self.files["Singleton"] = output_tmpl % ("singleton.truncated",)
            self.files["Paired"] = output_tmpl % ("pair{Pair}.truncated",)

            if collapse_reads:
                self.files["Collapsed"] = output_tmpl % ("collapsed",)
                self.files["CollapsedTruncated"] = output_tmpl % (
                    "collapsed.truncated",
                )

            init_args["collapse"] = collapse_reads
            init_args["input_file_1"] = record["Data"]["PE_1"]
            init_args["input_file_2"] = record["Data"]["PE_2"]
            command = PE_AdapterRemovalNode(**init_args)

        self.stats = os.path.join(self.folder, "reads.settings")
        self.nodes = (command,)



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Implements nodes for calling PALEOMIX commands.

Each node is equivalent to a particular command:
    $ paleomix [...]
"""
from paleomix.node import CommandNode, Node
from paleomix.atomiccmd.command import AtomicCmd
from paleomix.atomiccmd.sets import ParallelCmds
from paleomix.atomiccmd.builder import (
    AtomicCmdBuilder,
    apply_options,
)
from paleomix.common.fileutils import describe_files, reroot_path, move_file
from paleomix.nodes.samtools import merge_bam_files_command, BCFTOOLS_VERSION

import paleomix.tools.bam_stats.coverage as coverage
import paleomix.tools.factory as factory


class CoverageNode(CommandNode):
    def __init__(
        self, target_name, input_file, output_file, regions_file=None, dependencies=()
    ):
        builder = factory.new("coverage")
        builder.add_value("%(IN_BAM)s")
        builder.add_value("%(OUT_FILE)s")
        builder.set_option("--target-name", target_name)
        builder.set_kwargs(IN_BAM=input_file, OUT_FILE=output_file)

        if regions_file:
            builder.set_option("--regions-file", "%(IN_REGIONS)s")
            builder.set_kwargs(IN_REGIONS=regions_file)

        CommandNode.__init__(
            self,
            command=builder.finalize(),
            description="calculating coverage for %s" % (input_file,),
            dependencies=dependencies,
        )


class MergeCoverageNode(Node):
    def __init__(self, input_files, output_file, dependencies=()):
        self._output_file = output_file

        Node.__init__(
            self,
            description="merging coverage tables from %s"
            % (describe_files(input_files),),
            input_files=input_files,
            output_files=self._output_file,
            dependencies=dependencies,
        )

    def _run(self, _config, temp):
        table = {}
        for filename in self.input_files:
            coverage.read_table(table, filename)

        coverage.write_table(table, reroot_path(temp, self._output_file))
        move_file(reroot_path(temp, self._output_file), self._output_file)


class DepthHistogramNode(CommandNode):
    def __init__(
        self,
        target_name,
        input_file,
        output_file,
        prefix,
        regions_file=None,
        dependencies=(),
    ):
        index_format = regions_file and prefix["IndexFormat"]

        builder = factory.new("depths")
        builder.add_value("%(IN_BAM)s")
        builder.add_value("%(OUT_FILE)s")
        builder.set_option("--target-name", target_name)
        builder.set_kwargs(OUT_FILE=output_file, IN_BAM=input_file)

        if regions_file:
            builder.set_option("--regions-file", "%(IN_REGIONS)s")
            builder.set_kwargs(
                IN_REGIONS=regions_file,
                IN_INDEX=input_file + index_format,
            )

        CommandNode.__init__(
            self,
            command=builder.finalize(),
            description="calculating depth histogram for %s" % (input_file,),
            dependencies=dependencies,
        )


class FilterCollapsedBAMNode(CommandNode):
    def __init__(
        self, config, input_bams, output_bam, keep_dupes=True, dependencies=()
    ):
        merge = merge_bam_files_command(input_bams)

        builder = factory.new("rmdup_collapsed")
        builder.set_kwargs(IN_STDIN=merge, OUT_STDOUT=output_bam)

        if not keep_dupes:
            builder.set_option("--remove-duplicates")

        CommandNode.__init__(
            self,
            command=ParallelCmds([merge, builder.finalize()]),
            description="filtering collapsed PCR duplicates in %s"
            % (describe_files(merge.input_files),),
            dependencies=dependencies,
        )


class VCFFilterNode(CommandNode):
    def __init__(self, infile, outfile, regions, options, dependencies=()):
        vcffilter = factory.new("vcf_filter")
        vcffilter.add_value("%(IN_VCF)s")

        for contig in regions["HomozygousContigs"]:
            vcffilter.add_option("--homozygous-chromosome", contig)
        vcffilter.set_kwargs(IN_VCF=infile, OUT_STDOUT=AtomicCmd.PIPE)

        apply_options(vcffilter, options)

        bgzip = AtomicCmdBuilder(["bgzip"], IN_STDIN=vcffilter, OUT_STDOUT=outfile)

        CommandNode.__init__(
            self,
            description="filtering VCF records in %s" % (infile,),
            command=ParallelCmds([vcffilter.finalize(), bgzip.finalize()]),
            dependencies=dependencies,
        )


class GenotypeRegionsNode(CommandNode):
    def __init__(
        self,
        reference,
        infile,
        bedfile,
        outfile,
        mpileup_options={},
        bcftools_options={},
        dependencies=(),
    ):
        mpileup = AtomicCmdBuilder(
            ("bcftools", "mpileup", "%(IN_BAMFILE)s"),
            IN_BAMFILE=infile,
            IN_INTERVALS=bedfile,
            OUT_STDOUT=AtomicCmd.PIPE,
            CHECK_VERSION=BCFTOOLS_VERSION,
        )

        # Ignore read-groups for pileup
        mpileup.add_option("--ignore-RG")
        # Reference sequence (FASTA)
        mpileup.add_option("--fasta-ref", reference)
        # Output compressed VCF
        mpileup.add_option("--output-type", "u")

        if bedfile:
            mpileup.set_option("--regions-file", "%(IN_INTERVALS)s")

        apply_options(mpileup, mpileup_options)

        genotype = AtomicCmdBuilder(
            ("bcftools", "call", "-"),
            IN_STDIN=mpileup,
            IN_BAMFILE=infile,
            OUT_STDOUT=outfile,
            CHECK_VERSION=BCFTOOLS_VERSION,
        )

        genotype.set_option("--output-type", "z")

        apply_options(genotype, bcftools_options)

        CommandNode.__init__(
            self,
            description="calling genotypes from %s" % (infile,),
            command=ParallelCmds([mpileup.finalize(), genotype.finalize()]),
            dependencies=dependencies,
        )


class BuildRegionsNode(CommandNode):
    def __init__(self, infile, bedfile, outfile, padding, options={}, dependencies=()):
        params = factory.new("vcf_to_fasta")
        params.set_option("--padding", padding)
        params.set_option("--genotype", "%(IN_VCFFILE)s")
        params.set_option("--intervals", "%(IN_INTERVALS)s")

        params.set_kwargs(
            IN_VCFFILE=infile,
            IN_TABIX=infile + ".tbi",
            IN_INTERVALS=bedfile,
            OUT_STDOUT=outfile,
        )

        apply_options(params, options)

        CommandNode.__init__(
            self,
            description="building FASTA from %s" % (infile,),
            command=params.finalize(),
            dependencies=dependencies,
        )


def _apply_samtools_options(builder, options, argument):
    for (key, value) in dict(options).items():
        sam_argument = key
        if value is not None:
            sam_argument = "%s=%s" % (key, value)

        builder.add_option(argument, sam_argument, sep="=")



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os
import signal
import weakref

from unittest.mock import call, Mock, patch

import pytest

import paleomix.atomiccmd.command
import paleomix.common.fileutils as fileutils

from paleomix.common.versions import RequirementObj
from paleomix.atomiccmd.command import AtomicCmd, CmdError


###############################################################################
###############################################################################
# Constructor: Command


def test_atomiccmd__command_str():
    cmd = AtomicCmd("ls")
    assert cmd.executables == frozenset(["ls"])


def test_atomiccmd__executables_empty_str():
    with pytest.raises(ValueError, match="Empty command in AtomicCmd constructor"):
        AtomicCmd("")


def test_atomiccmd__command_tuple():
    cmd = AtomicCmd(("cd", "."))
    assert cmd.executables == frozenset(["cd"])


def test_atomiccmd__executables_empty_tuple():
    with pytest.raises(ValueError, match="Empty command in AtomicCmd constructor"):
        AtomicCmd(())


def test_atomiccmd__executables_empty_str_in_tuple():
    with pytest.raises(ValueError, match="Empty command in AtomicCmd constructor"):
        AtomicCmd((""))


###############################################################################
###############################################################################
# Constructor: set_cwd


@pytest.mark.parametrize("set_cwd", (True, False))
def test_atomiccmd__set_cwd(tmp_path, set_cwd):
    cwd = os.getcwd()
    cmd = AtomicCmd(
        ("bash", "-c", "echo -n ${PWD}"), TEMP_OUT_STDOUT="result.txt", set_cwd=set_cwd
    )
    cmd.run(tmp_path)
    assert cmd.join() == [0]
    assert cwd == os.getcwd()

    expected = str(tmp_path) if set_cwd else cwd
    result = (tmp_path / "result.txt").read_text()
    assert os.path.samefile(expected, result), "%r != %r" % (expected, result)


# Full path when set_cwd is False, rel. path when True
@pytest.mark.parametrize("set_cwd", (True, False))
@pytest.mark.parametrize("key", ("TEMP_IN_FOO", "TEMP_OUT_FOO"))
def test_atomiccmd__set_cwd__temp_in_out(tmp_path, set_cwd, key):
    cmd = AtomicCmd(
        ("echo", "-n", "%%(%s)s" % (key,)),
        TEMP_OUT_STDOUT="result.txt",
        set_cwd=set_cwd,
        **{key: "test_file"},
    )
    cmd.run(tmp_path)
    assert cmd.join() == [0]

    expected = os.path.join("" if set_cwd else str(tmp_path), "test_file")
    result = (tmp_path / "result.txt").read_text()
    assert os.path.abspath(expected) == os.path.abspath(result)


###############################################################################
###############################################################################
# Constructor: Paths / pipes

# Check that specified paths/etc. are available via getters
def test_atomiccmd__paths():
    cmd = AtomicCmd(
        "ls",
        IN_AAA="/a/b/c",
        IN_AAB="/x/y/z",
        TEMP_IN_ABB="tmp_in",
        OUT_AAA="/out/foo",
        OUT_BBC="foo/bar",
        TEMP_OUT_A="xyb",
        EXEC_OTHER="true",
        AUX_WAT="wat/wat",
        CHECK_FUNC=bool,
        OUT_STDERR="/var/log/pipe.stderr",
        TEMP_OUT_STDOUT="pipe.stdout",
    )

    assert cmd.executables == frozenset(["ls", "true"])
    assert cmd.requirements == frozenset([bool])
    assert cmd.input_files == frozenset(["/a/b/c", "/x/y/z"])
    assert cmd.output_files == frozenset(
        ["/out/foo", "foo/bar", "/var/log/pipe.stderr"]
    )
    assert cmd.auxiliary_files == frozenset(["wat/wat"])
    assert cmd.expected_temp_files == frozenset(["foo", "bar", "pipe.stderr"])
    assert "xyb" in cmd.optional_temp_files
    assert "pipe.stdout" in cmd.optional_temp_files


def test_atomiccmd__paths_optional():
    cmd = AtomicCmd(["ls"], IN_OPTIONAL=None, OUT_OPTIONAL=None)
    assert cmd.input_files == frozenset()
    assert cmd.output_files == frozenset()


def test_atomiccmd__pipes_stdin(tmp_path):
    fname = tmp_path / "input.fasta"
    fname.write_text(">This_is_FASTA!\nACGTN\n>This_is_ALSO_FASTA!\nCGTNA\n")
    fname = str(fname)

    cmd = AtomicCmd("cat", IN_STDIN=fname, OUT_STDOUT="result.txt")
    assert cmd.input_files == frozenset([fname])
    cmd.run(tmp_path)
    assert cmd.join() == [0]
    result = (tmp_path / "result.txt").read_text()
    assert result == ">This_is_FASTA!\nACGTN\n>This_is_ALSO_FASTA!\nCGTNA\n"


def test_atomiccmd__pipes_stdin__temp_file(tmp_path):
    cmd = AtomicCmd("cat", TEMP_IN_STDIN="infile.fasta", OUT_STDOUT="result.txt")
    assert cmd.input_files == frozenset()
    (tmp_path / "infile.fasta").write_text("a\nbc\nd")
    cmd.run(tmp_path)
    assert cmd.join() == [0]
    result = (tmp_path / "result.txt").read_text()
    assert result == "a\nbc\nd"


def test_atomiccmd__pipes_stdin__dev_null_implicit_1(tmp_path):
    # STDIN should be implicitly set to /dev/null; deadlocks if not
    cmd = AtomicCmd("cat")
    cmd.run(tmp_path)
    assert cmd.join() == [0]


def test_atomiccmd__pipes_stdin__dev_null_implicit_2(tmp_path):
    # STDIN should be implicitly set to /dev/null; deadlocks if not
    cmd = AtomicCmd("cat", IN_STDIN=None)
    cmd.run(tmp_path)
    assert cmd.join() == [0]


def test_atomiccmd__pipes_stdin__dev_null_explicit(tmp_path):
    # STDIN should be set to /dev/null; deadlocks if not
    cmd = AtomicCmd("cat", IN_STDIN=AtomicCmd.DEVNULL)
    cmd.run(tmp_path, wrap_errors=False)
    assert cmd.join() == [0]


# Test possible combinations of explicit / implicit saving of stdout/err
_TEST_OUTPUT_PIPES = (
    ("pipe_bash_{0}.stdout", "pipe_bash_{0}.stderr", {}),
    ("pipe_bash_{0}.stdout", "stderr.txt", {"OUT_STDERR": "stderr.txt"}),
    ("stdout.txt", "pipe_bash_{0}.stderr", {"OUT_STDOUT": "stdout.txt"}),
    (
        "stdout.txt",
        "stderr.txt",
        {"OUT_STDOUT": "stdout.txt", "OUT_STDERR": "stderr.txt"},
    ),
    (None, None, {"OUT_STDOUT": AtomicCmd.DEVNULL, "OUT_STDERR": AtomicCmd.DEVNULL}),
    (None, "pipe_bash_{0}.stderr", {"OUT_STDOUT": AtomicCmd.DEVNULL}),
    ("pipe_bash_{0}.stdout", None, {"OUT_STDERR": AtomicCmd.DEVNULL}),
)


@pytest.mark.parametrize("stdout, stderr, kwargs", _TEST_OUTPUT_PIPES)
def test_atomiccmd__pipes_out(tmp_path, stdout, stderr, kwargs):
    call = ("bash", "-c", "echo -n 'STDERR!' > /dev/stderr; echo -n 'STDOUT!';")

    cmd = AtomicCmd(call, **kwargs)
    cmd.run(tmp_path)
    assert cmd.join() == [0]

    expected_files = []
    for (tmpl, text) in ((stdout, "STDOUT!"), (stderr, "STDERR!")):
        if tmpl is not None:
            fname = tmpl.format(id(cmd))
            result = (tmp_path / fname).read_text()
            assert result == text
            expected_files.append(fname)

    assert set(os.listdir(str(tmp_path))) == set(expected_files)


_MALFORMED_PATH_KEYS = (
    {"IN": "/var/foo"},  # Missing key-name #1
    {"IN_": "/var/foo"},  # Missing key-name #2
    {"TEMP_OUT": "/var/foo"},  # Missing key-name #3
    {"TEMP_OUT_": "/var/foo"},  # Missing key-name #4
    {"TEMP_OUX_FOO": "foo"},  # Invalid key-type #1
    {"INS_BAR": "foo"},  # Invalid key-type #2
)


@pytest.mark.parametrize("kwargs", _MALFORMED_PATH_KEYS)
def test_atomiccmd__paths__malformed_keys(kwargs):
    with pytest.raises(ValueError):
        AtomicCmd("true", **kwargs)


_INVALID_PATH_VALUES = (
    {"IN_FILE": 1},
    {"TEMP_IN_FILE": set()},
    {"OUT_FILE": [1, 2, 3]},
    {"TEMP_OUT_FILE": 1.0},
    {"IN_STDIN": {}},
    {"TEMP_IN_STDIN": frozenset()},
    {"OUT_STDOUT": 1.7},
    {"TEMP_OUT_STDOUT": ()},
    {"OUT_STDERR": range(3)},
    {"TEMP_OUT_STDERR": -1},
)


@pytest.mark.parametrize("kwargs", _INVALID_PATH_VALUES)
def test_atomiccmd__paths__invalid_values(kwargs):
    with pytest.raises(TypeError):
        AtomicCmd("true", **kwargs)


# Subpaths are not allowed for temp IN/OUT files, neither relative nor asbsolute
_INVALID_TEMP_PATHS = (
    # No relative paths
    {"TEMP_IN_FOO": "sub/infile"},
    {"TEMP_IN_STDIN": "sub/stdin"},
    {"TEMP_OUT_FOO": "sub/outfile"},
    {"TEMP_OUT_STDOUT": "sub/stdout"},
    {"TEMP_OUT_STDERR": "sub/stderr"},
    # No absolute paths
    {"TEMP_IN_FOO": "/tmp/sub/infile"},
    {"TEMP_IN_STDIN": "/dev/sub/stdin"},
    {"TEMP_OUT_FOO": "/etc/sub/outfile"},
    {"TEMP_OUT_STDOUT": "/var/sub/stdout"},
    {"TEMP_OUT_STDERR": "/home/sub/stderr"},
)


@pytest.mark.parametrize("kwargs", _INVALID_TEMP_PATHS)
def test_atomiccmd__paths__invalid_temp_paths(kwargs):
    with pytest.raises(ValueError):
        AtomicCmd("true", **kwargs)


_OVERLAPPING_OUT_FILENAMES = (
    {"OUT_FILE_1": "/foo/bar/outfile", "OUT_FILE_2": "/var/outfile"},
    {"TEMP_OUT_FILE_1": "outfile", "OUT_FILE_1": "/var/outfile"},
    {"OUT_FILE_1": "/foo/bar/outfile", "TEMP_OUT_FILE_1": "outfile"},
    {"TEMP_OUT_FILE_1": "outfile", "TEMP_OUT_FILE_2": "outfile"},
    {"OUT_FILE_1": "/foo/bar/outfile", "OUT_STDOUT": "/var/outfile"},
    {"TEMP_OUT_FILE_1": "outfile", "OUT_STDOUT": "/var/outfile"},
    {"OUT_FILE_1": "/foo/bar/outfile", "TEMP_OUT_STDOUT": "outfile"},
    {"TEMP_OUT_FILE_1": "outfile", "TEMP_OUT_STDOUT": "outfile"},
    {"OUT_FILE_1": "/foo/bar/outfile", "OUT_STDERR": "/var/outfile"},
    {"TEMP_OUT_FILE_1": "outfile", "OUT_STDERR": "/var/outfile"},
    {"OUT_FILE_1": "/foo/bar/outfile", "TEMP_OUT_STDERR": "outfile"},
    {"TEMP_OUT_FILE_1": "outfile", "TEMP_OUT_STDERR": "outfile"},
)

# All OUT_ files must be unique, including all TEMP_OUT_
@pytest.mark.parametrize("kwargs", _OVERLAPPING_OUT_FILENAMES)
def test_atomiccmd__paths__overlapping_output(kwargs):
    with pytest.raises(ValueError):
        AtomicCmd(("ls",), **kwargs)


# A pipe can be w/wo TEMP_, but not both
@pytest.mark.parametrize("key", ("IN_STDIN", "OUT_STDOUT", "OUT_STDERR"))
def test_atomiccmd__pipes__duplicates(key):
    kwargs = {"TEMP_" + key: "temp_file", key: "file"}
    with pytest.raises(CmdError):
        AtomicCmd(["ls"], **kwargs)


###############################################################################
###############################################################################
# CHECK_ / EXEC_

# RequirementObjs are the standard way to do tests
def test_atomicmcd__exec__reqobj():
    reqobj = RequirementObj(call=("echo", "version"), search="version", checks=str)
    cmd = AtomicCmd("true", CHECK_VERSION=reqobj)
    assert cmd.requirements == frozenset([reqobj])


# CHECK_ is expected to be a callable
def test_atomiccmd__checks__non_callable():
    with pytest.raises(TypeError, match="CHECK_FOO must be callable"):
        AtomicCmd("ls", CHECK_FOO="ls")


# EXEC_ is expected to be a string
@pytest.mark.parametrize("obj", (str, {}, 1, b"foo"))
def test_atomiccmd__exec__invalid(obj):
    with pytest.raises(TypeError, match="EXEC_FOO must be string, not "):
        AtomicCmd("true", EXEC_FOO=obj)


###############################################################################
###############################################################################
# AUX


@pytest.mark.parametrize("obj", (str, {}, 1, b"foo"))
def test_atomiccmd__aux__invalid(obj):
    with pytest.raises(TypeError, match="AUX_FOO must be string, not "):
        AtomicCmd("true", AUX_FOO=obj)


###############################################################################
###############################################################################
# Path components


def test_atomiccmd__paths_non_str(tmp_path):
    cmd = AtomicCmd(("touch", 1234), OUT_FOO="1234", set_cwd=True)
    cmd.run(tmp_path)
    assert cmd.join() == [0]
    assert (tmp_path / "1234").exists()


def test_atomiccmd__paths_missing():
    with pytest.raises(CmdError, match="Value not specified for path"):
        AtomicCmd(("touch", "%(IN_FOO)s"))


def test_atomiccmd__paths_invalid():
    with pytest.raises(CmdError, match="incomplete format"):
        AtomicCmd(("touch", "%(IN_FOO)"), IN_FOO="abc")


def test_atomiccmd__paths__key(tmp_path):
    cmd = AtomicCmd(("echo", "-n", "%(TEMP_DIR)s"), OUT_STDOUT=AtomicCmd.PIPE)
    cmd.run(tmp_path)
    path = cmd._proc.stdout.read()
    assert tmp_path.samefile(path), (tmp_path, path)
    assert cmd.join() == [0]


###############################################################################
###############################################################################
# Constructor: Piping commands


def test_atomiccmd__piping(tmp_path):
    cmd_1 = AtomicCmd(["echo", "-n", "#@!$^"], OUT_STDOUT=AtomicCmd.PIPE)
    assert cmd_1.output_files == frozenset()
    cmd_2 = AtomicCmd(["cat"], IN_STDIN=cmd_1, OUT_STDOUT="piped.txt")
    assert cmd_2.input_files == frozenset()
    cmd_1.run(tmp_path)
    cmd_2.run(tmp_path)
    assert cmd_1.join() == [0]
    assert cmd_2.join() == [0]
    result = (tmp_path / "piped.txt").read_text()
    assert result == "#@!$^"


def test_atomiccmd__piping_temp(tmp_path):
    cmd_1 = AtomicCmd(["echo", "-n", "#@!$^"], TEMP_OUT_STDOUT=AtomicCmd.PIPE)
    assert cmd_1.output_files == frozenset()
    cmd_2 = AtomicCmd(["cat"], TEMP_IN_STDIN=cmd_1, OUT_STDOUT="piped.txt")
    assert cmd_2.input_files == frozenset()
    cmd_1.run(tmp_path)
    cmd_2.run(tmp_path)
    assert cmd_1.join() == [0]
    assert cmd_2.join() == [0]
    result = (tmp_path / "piped.txt").read_text()
    assert result == "#@!$^"


# Only STDOUT takes AtomicCmd.PIPE
_INVALID_PIPE_TARGETS = (
    "IN_FILE_1",
    "IN_STDIN",
    "OUT_FILE_2",
    "OUT_STDERR",
    "TEMP_IN_FILE_1",
    "TEMP_IN_STDIN",
    "TEMP_OUT_FILE_2",
    "TEMP_OUT_STDERR",
)


@pytest.mark.parametrize("key", _INVALID_PIPE_TARGETS)
def test_atomiccmd__piping__wrong_pipe(key):
    with pytest.raises(TypeError):
        AtomicCmd("ls", **{key: AtomicCmd.PIPE})


def test_atomiccmd__piping_is_only_allowed_once(tmp_path):
    cmd_1 = AtomicCmd(["echo", "-n", "foo\nbar"], OUT_STDOUT=AtomicCmd.PIPE)
    cmd_2a = AtomicCmd(["grep", "foo"], IN_STDIN=cmd_1)
    cmd_2b = AtomicCmd(["grep", "bar"], IN_STDIN=cmd_1)
    cmd_1.run(tmp_path)
    cmd_2a.run(tmp_path)
    with pytest.raises(CmdError):
        cmd_2b.run(tmp_path)
    assert cmd_1.join() == [0]
    assert cmd_2a.join() == [0]
    assert cmd_2b.join() == [None]


###############################################################################
###############################################################################
# run


def test_atomiccmd__run__already_running(tmp_path):
    cmd = AtomicCmd(("sleep", "10"))
    cmd.run(tmp_path)
    with pytest.raises(CmdError):
        cmd.run(tmp_path)
    cmd.terminate()
    cmd.join()


def test_atomiccmd__run__exception_on_missing_command(tmp_path):
    cmd = AtomicCmd(("xyzabcefgh", "10"))
    with pytest.raises(CmdError):
        cmd.run(tmp_path)
    cmd.terminate()
    cmd.join()


def test_atomiccmd__run__exception_on_missing_command__no_wrap(tmp_path):
    cmd = AtomicCmd(("xyzabcefgh", "10"))
    with pytest.raises(OSError):
        cmd.run(tmp_path, wrap_errors=False)
    cmd.terminate()
    cmd.join()


def test_atomiccmd__run__invalid_temp(tmp_path):
    cmd = AtomicCmd(("sleep", "10"))
    with pytest.raises(CmdError):
        cmd.run(tmp_path / "foo")
    cmd.terminate()
    cmd.join()


###############################################################################
###############################################################################
# Ready


def test_atomiccmd__ready(tmp_path):
    cmd = AtomicCmd("ls")
    assert cmd.join() == [None]
    assert not cmd.ready()
    cmd.run(tmp_path)
    assert cmd.join() == [0]
    assert cmd.ready()


###############################################################################
###############################################################################
# Join / wait


@pytest.mark.parametrize("call, after", (("true", 0), ("false", 1)))
def test_atomiccmd__join(tmp_path, call, after):
    cmd = AtomicCmd(call)
    assert cmd.join() == [None]
    cmd.run(tmp_path)
    assert cmd.join() == [after]


@pytest.mark.parametrize("call, after", (("true", 0), ("false", 1)))
def test_atomiccmd__wait(tmp_path, call, after):
    cmd = AtomicCmd(call)
    assert cmd.wait() is None
    cmd.run(tmp_path)
    assert cmd.wait() == after


###############################################################################
###############################################################################
# Terminate


def test_atomiccmd__terminate(tmp_path):
    cmd = AtomicCmd(("sleep", "10"))
    cmd.run(tmp_path)

    with patch("os.killpg", wraps=os.killpg) as os_killpg:
        cmd.terminate()
        assert cmd.join() == ["SIGTERM"]

        assert os_killpg.mock_calls == [call(cmd._proc.pid, signal.SIGTERM)]


def test_atomiccmd__terminate_exception(tmp_path):
    killpg = os.killpg
    cmd = AtomicCmd(("sleep", "10"))
    cmd.run(tmp_path)

    def _killpg(pid, sig):
        killpg(pid, sig)
        raise OSError("Proccess not found")

    with patch("os.killpg", wraps=_killpg) as os_killpg:
        cmd.terminate()
        assert cmd.join() == ["SIGTERM"]

        assert os_killpg.mock_calls == [call(cmd._proc.pid, signal.SIGTERM)]


# Ensure that no OSException is raised, even if the command
# managed to finish before terminate was called
def test_atomiccmd__terminate_race_condition(tmp_path):
    cmd = AtomicCmd("true")
    cmd.run(tmp_path)
    while cmd._proc.poll() is None:
        pass
    cmd.terminate()
    assert cmd.join() == [0]


# Calling terminate on an already joined command is acceptable
def test_atomiccmd__terminate_after_join(tmp_path):
    cmd = AtomicCmd("true")
    cmd.run(tmp_path)
    assert cmd.join() == [0]
    cmd.terminate()
    assert cmd.join() == [0]


# Signals are translated into strings
def test_atomiccmd__terminate_sigterm(tmp_path):
    cmd = AtomicCmd(("sleep", "10"))
    cmd.run(tmp_path)
    cmd.terminate()
    assert cmd.join() == ["SIGTERM"]


def test_atomiccmd__terminate_sigkill(tmp_path):
    cmd = AtomicCmd(("sleep", "10"))
    cmd.run(tmp_path)
    cmd._proc.kill()
    assert cmd.join() == ["SIGKILL"]


###############################################################################
###############################################################################
# commit


def _setup_for_commit(tmp_path, create_cmd=True):
    destination = tmp_path / "out"
    tmp_path = tmp_path / "tmp"
    destination.mkdir(parents=True)
    tmp_path.mkdir(parents=True)

    if not create_cmd:
        return destination, tmp_path

    cmd = AtomicCmd(("touch", "%(OUT_FOO)s"), OUT_FOO=str(destination / "1234"))
    cmd.run(tmp_path)
    assert cmd.join() == [0]

    return destination, tmp_path, cmd


def test_atomiccmd__commit_simple(tmp_path):
    destination, tmp_path, cmd = _setup_for_commit(tmp_path)
    cmd.commit(tmp_path)
    assert not (tmp_path / "1234").exists()
    assert (destination / "1234").exists()


def test_atomiccmd__commit_temp_out(tmp_path):
    dest, temp = _setup_for_commit(tmp_path, create_cmd=False)
    cmd = AtomicCmd(
        ("echo", "foo"), OUT_STDOUT=str(dest / "foo.txt"), TEMP_OUT_FOO="bar.txt",
    )
    cmd.run(temp)
    assert cmd.join() == [0]
    (temp / "bar.txt").write_text("1 2 3")
    cmd.commit(temp)
    assert os.listdir(str(temp)) == []
    assert os.listdir(str(dest)) == ["foo.txt"]


def test_atomiccmd__commit_temp_only(tmp_path):
    cmd = AtomicCmd(("echo", "foo"), TEMP_OUT_STDOUT="bar.txt")
    cmd.run(tmp_path)
    assert cmd.join() == [0]
    assert (tmp_path / "bar.txt").exists()
    cmd.commit(tmp_path)
    assert os.listdir(str(tmp_path)) == []


def test_atomiccmd__commit_before_run():
    cmd = AtomicCmd("true")
    with pytest.raises(CmdError):
        cmd.commit("/tmp")


def test_atomiccmd__commit_while_running(tmp_path):
    cmd = AtomicCmd(("sleep", "10"))
    cmd.run(tmp_path)
    with pytest.raises(CmdError):
        cmd.commit(tmp_path)
    cmd.terminate()
    cmd.join()


def test_atomiccmd__commit_before_join(tmp_path):
    cmd = AtomicCmd(("sleep", "0.1"))
    cmd.run(tmp_path)
    while cmd._proc.poll() is None:
        pass
    with pytest.raises(CmdError):
        cmd.commit(tmp_path)
    cmd.join()


# The temp path might differ, as long as the actual path is the same
def test_atomiccmd__commit_temp_folder(tmp_path):
    destination, tmp_path, cmd = _setup_for_commit(tmp_path)
    cmd.commit(tmp_path.resolve())
    assert not (tmp_path / "1234").exists()
    assert (destination / "1234").exists()


def test_atomiccmd__commit_wrong_temp_folder(tmp_path):
    destination, tmp_path, cmd = _setup_for_commit(tmp_path)
    with pytest.raises(CmdError):
        cmd.commit(destination)


def test_atomiccmd__commit_missing_files(tmp_path):
    destination, tmp_path = _setup_for_commit(tmp_path, False)
    cmd = AtomicCmd(
        ("touch", "%(OUT_FOO)s"),
        OUT_FOO=str(destination / "1234"),
        OUT_BAR=str(destination / "4567"),
    )
    cmd.run(tmp_path)
    cmd.join()
    before = frozenset(tmp_path.iterdir())
    with pytest.raises(CmdError):
        cmd.commit(tmp_path)
    assert before == frozenset(tmp_path.iterdir())


def test_atomiccmd__commit_failure_cleanup(tmp_path):
    counter = []
    move_file = fileutils.move_file

    def _monkey_move_file(source, destination):
        if counter:
            raise OSError("ARRRGHHH!")
        counter.append(destination)

        return move_file(source, destination)

    destination, tmp_path = _setup_for_commit(tmp_path, False)
    command = AtomicCmd(
        ("touch", "%(OUT_FILE_1)s", "%(OUT_FILE_2)s", "%(OUT_FILE_3)s"),
        OUT_FILE_1=str(destination / "file_1"),
        OUT_FILE_2=str(destination / "file_2"),
        OUT_FILE_3=str(destination / "file_3"),
    )

    try:
        fileutils.move_file = _monkey_move_file
        command.run(tmp_path)
        assert command.join() == [0]
        with pytest.raises(OSError):
            command.commit(tmp_path)

        assert tuple(destination.iterdir()) == ()
    finally:
        fileutils.move_file = move_file


def test_atomiccmd__commit_with_pipes(tmp_path):
    destination, tmp_path = _setup_for_commit(tmp_path, False)
    command_1 = AtomicCmd(("echo", "Hello, World!"), OUT_STDOUT=AtomicCmd.PIPE)
    command_2 = AtomicCmd(
        ("gzip",), IN_STDIN=command_1, OUT_STDOUT=str(destination / "foo.gz")
    )

    command_1.run(tmp_path)
    command_2.run(tmp_path)

    assert command_1.join() == [0]
    assert command_2.join() == [0]

    command_1.commit(tmp_path)
    command_2.commit(tmp_path)

    assert list(destination.iterdir()) == [destination / "foo.gz"]
    assert list(tmp_path.iterdir()) == []


###############################################################################
###############################################################################
# __str__
# Additional tests in atomicpp_test.py


def test_atomiccmd__str__():
    cmd = AtomicCmd(("echo", "test"))
    assert paleomix.atomiccmd.pprint.pformat(cmd) == str(cmd)


###############################################################################
###############################################################################
# Cleanup

# FIXME: Needs better tracking of procs
#        1. track for termination until joined
#        2. don't use weak references to avoid accidental leaks

# Test that the internal list of processes is kept clean of old objects
def test_atomiccmd__cleanup_proc__commit(tmp_path):
    assert paleomix.atomiccmd.command._PROCS == set()
    cmd = AtomicCmd("ls")
    cmd.run(tmp_path)

    ref = next(iter(paleomix.atomiccmd.command._PROCS))
    assert ref() == cmd._proc
    assert cmd.join() == [0]
    # Commit frees proc object
    cmd.commit(tmp_path)
    assert ref() is None

    assert ref not in paleomix.atomiccmd.command._PROCS


# Test that the internal list of processes is kept clean of old objects
def test_atomiccmd__cleanup_proc__gc(tmp_path):
    assert paleomix.atomiccmd.command._PROCS == set()
    cmd = AtomicCmd("ls")
    cmd.run(tmp_path)

    ref = next(iter(paleomix.atomiccmd.command._PROCS))
    assert ref() == cmd._proc
    assert cmd.join() == [0]
    # GC frees proc object
    cmd = None
    assert ref() is None

    assert ref not in paleomix.atomiccmd.command._PROCS


def test_atomiccmd__cleanup_sigterm():
    procs = [lambda: Mock(pid=7913), lambda: Mock(pid=12345)]
    with patch("paleomix.atomiccmd.command._PROCS", procs):
        patches = Mock()
        with patch("os.killpg", patches.killpg):
            with patch("sys.exit", patches.exit):
                paleomix.atomiccmd.command._cleanup_children(signal.SIGTERM, None)

                assert patches.mock_calls == [
                    call.killpg(7913, signal.SIGTERM),
                    call.killpg(12345, signal.SIGTERM),
                    call.exit(-signal.SIGTERM),
                ]


def test_atomiccmd__cleanup_sigterm__continues_on_exception():
    procs = [lambda: Mock(pid=7913), lambda: Mock(pid=12345)]
    with patch("paleomix.atomiccmd.command._PROCS", procs):
        patches = Mock()
        with patch("os.killpg", patches.killpg):
            with patch("sys.exit", patches.exit):
                patches.killpg.side_effect = [OSError("already killed"), None]

                paleomix.atomiccmd.command._cleanup_children(signal.SIGTERM, None)

                assert patches.mock_calls == [
                    call.killpg(7913, signal.SIGTERM),
                    call.killpg(12345, signal.SIGTERM),
                    call.exit(-signal.SIGTERM),
                ]


# Ensure that the cleanup function handles weakrefs that have been freed
def test_atomiccmd__cleanup_sigterm__dead_weakrefs():
    dead_ref = weakref.ref(AtomicCmd("ls"))
    with patch("paleomix.atomiccmd.command._PROCS", [dead_ref]):
        patches = Mock()
        with patch("os.killpg", patches.killpg):
            with patch("sys.exit", patches.exit):
                paleomix.atomiccmd.command._cleanup_children(signal.SIGTERM, None)

                assert patches.mock_calls == [call.exit(-signal.SIGTERM)]



#!/usr/bin/python3
#
# Copyright (c) 2016 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import paleomix.common.versions as versions

from paleomix.resources import rscript


def requirement(module, checks=versions.Any(), cache={}):
    key = (module, checks)
    result = cache.get(key)
    if result is None:
        filename = rscript("common", "requires.r")
        result = versions.Requirement(
            call=("Rscript", filename, module),
            search=r"d0fd3ea6: (\d+)\.(\d+)(?:\.(\d+))?",
            checks=checks,
            name="R module: {}".format(module),
        )

        cache[(module, checks)] = result

    return result



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os
import re
import random

import paleomix.common.fileutils as fileutils
import paleomix.common.versions as versions

from paleomix.node import CommandNode
from paleomix.atomiccmd.builder import AtomicCmdBuilder


RAXML_VERSION = versions.Requirement(
    call=("raxmlHPC", "-version"),
    search=r"version (\d+)\.(\d+)\.(\d+)",
    checks=versions.GE(8, 2, 9),
)
RAXML_PTHREADS_VERSION = versions.Requirement(
    call=("raxmlHPC-PTHREADS", "-version"),
    search=r"version (\d+)\.(\d+)\.(\d+)",
    checks=versions.GE(8, 2, 9),
)


class RAxMLRapidBSNode(CommandNode):
    def __init__(
        self,
        input_alignment,
        output_template,
        input_partition=None,
        model="GTRGAMMAI",
        replicates="autoMRE",
        threads=1,
        dependencies=(),
    ):
        """
        Arguments:
        input_alignment  -- An alignment file in a format readable by RAxML.
        input_partition  -- A set of partitions in a format readable by RAxML.
        output_template  -- A template string used to construct final filenames. Should
                            consist of a full path, including a single '%s', which is
                            replaced with the  variable part of RAxML output files (e.g.
                            'info', 'bestTree', ...).

                            Example destination: '/disk/project/SN013420.RAxML.%s'
                            Example output:      '/disk/project/SN013420.RAxML.bestTree'
        """

        if threads > 1:
            command = AtomicCmdBuilder("raxmlHPC-PTHREADS")
            command.set_option("-T", threads)
            version = RAXML_PTHREADS_VERSION
        else:
            command = AtomicCmdBuilder("raxmlHPC")
            version = RAXML_VERSION

        # Perform rapid bootstrapping
        command.set_option("-f", "a")
        # Output files are saved with a .PALEOMIX postfix, and subsequently renamed
        command.set_option("-n", "PALEOMIX")
        # Ensures that output is saved to the temporary directory
        command.set_option("-w", "%(TEMP_DIR)s")
        # Symlink to sequence and partitions, to prevent the creation of *.reduced files
        # outside temp folder. In addition, it may be nessesary to remove the .reduced
        # files if created
        command.set_option("-s", "%(TEMP_OUT_ALN)s")

        if input_partition is not None:
            command.set_option("-q", "%(TEMP_OUT_PART)s")
            command.set_kwargs(
                IN_PARTITION=input_partition,
                TEMP_OUT_PART=os.path.basename(input_partition),
                TEMP_OUT_PART_R=os.path.basename(input_partition) + ".reduced",
            )

        command.set_kwargs(
            # Auto-delete: Symlinks and .reduced files that RAxML may generate
            TEMP_OUT_ALN=os.path.basename(input_alignment),
            TEMP_OUT_ALN_R=os.path.basename(input_alignment) + ".reduced",
            # Input files, are not used directly (see below)
            IN_ALIGNMENT=input_alignment,
            # Final output files, are not created directly
            OUT_INFO=output_template % "info",
            OUT_BESTTREE=output_template % "bestTree",
            OUT_BOOTSTRAP=output_template % "bootstrap",
            OUT_BIPART=output_template % "bipartitions",
            OUT_BIPARTLABEL=output_template % "bipartitionsBranchLabels",
            CHECK_VERSION=version,
        )

        # Use the GTRGAMMA model of NT substitution by default
        command.set_option("-m", model, fixed=False)
        # Enable Rapid Boostrapping and set random seed. May be set to a fixed value to
        # allow replicability.
        command.set_option("-x", int(random.random() * 2 ** 31 - 1), fixed=False)
        # Set random seed for parsimony inference. May be set to allow replicability.
        command.set_option("-p", int(random.random() * 2 ** 31 - 1), fixed=False)
        # Terminate bootstrapping upon convergence, not after N repetitions
        command.set_option("-N", replicates, fixed=False)

        self._symlinks = [input_alignment, input_partition]
        self._template = os.path.basename(output_template)

        CommandNode.__init__(
            self,
            command=command.finalize(),
            description="inferring phylogeny from %s using RAxML" % (input_alignment,),
            threads=threads,
            dependencies=dependencies,
        )

    def _setup(self, config, temp):
        CommandNode._setup(self, config, temp)

        # Required to avoid the creation of files outside the temp folder
        for filename in self._symlinks:
            if filename is not None:
                source = os.path.abspath(filename)
                destination = os.path.join(temp, os.path.basename(filename))

                os.symlink(source, destination)

    def _teardown(self, config, temp):
        for filename in os.listdir(temp):
            match = re.match("RAxML_(.*).PALEOMIX", filename)
            if match:
                source = os.path.join(temp, filename)
                destination = os.path.join(temp, self._template % match.groups())

                fileutils.move_file(source, destination)

        CommandNode._teardown(self, config, temp)


class RAxMLParsimonyTreeNode(CommandNode):
    def __init__(self, input_alignment, input_partitions, output_tree, dependencies=()):
        command = AtomicCmdBuilder("raxmlHPC")

        # Compute a randomized parsimony starting tree
        command.set_option("-y")
        # Output files are saved with a .Pypeline postfix, and subsequently renamed
        command.set_option("-n", "Pypeline")
        # Model required, but not used
        command.set_option("-m", "GTRGAMMA")
        # Ensures that output is saved to the temporary directory
        command.set_option("-w", "%(TEMP_DIR)s")
        # Set random seed for bootstrap generation. May be set to allow replicability
        command.set_option("-p", int(random.random() * 2 ** 31 - 1), fixed=False)

        # Symlink to sequence and partitions, to prevent the creation of *.reduced files
        # outside temp folder
        command.set_option("-s", "%(TEMP_OUT_ALIGNMENT)s")
        command.set_option("-q", "%(TEMP_OUT_PARTITION)s")

        command.set_kwargs(
            IN_ALIGNMENT=input_alignment,
            IN_PARTITION=input_partitions,
            # TEMP_OUT_ is used to automatically remove these files
            TEMP_OUT_ALIGNMENT="RAxML_alignment",
            TEMP_OUT_PARTITION="RAxML_partitions",
            TEMP_OUT_INFO="RAxML_info.Pypeline",
            OUT_TREE=output_tree,
            CHECK_VERSION=RAXML_VERSION,
        )

        self._input_alignment = input_alignment
        self._input_partitions = input_partitions
        self._output_tree = output_tree

        CommandNode.__init__(
            self,
            command=command.finalize(),
            description="inferring parsimony phylogeny from %s using RAxML"
            % (input_alignment,),
            dependencies=dependencies,
        )

    def _setup(self, config, temp):
        os.symlink(
            os.path.abspath(self._input_alignment),
            os.path.join(temp, "RAxML_alignment"),
        )
        os.symlink(
            os.path.abspath(self._input_partitions),
            os.path.join(temp, "RAxML_partitions"),
        )
        CommandNode._setup(self, config, temp)

    def _teardown(self, config, temp):
        basename = os.path.basename(self._output_tree)
        os.rename(
            os.path.join(temp, "RAxML_parsimonyTree.Pypeline"),
            os.path.join(temp, basename),
        )

        CommandNode._teardown(self, config, temp)



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import collections

import pytest

from paleomix.common.text import (
    TableError,
    padded_table,
    parse_lines,
    parse_lines_by_contig,
    parse_padded_table,
)

###############################################################################
###############################################################################
# Tests for 'padded_table'


def _padded_table(*args, **kwargs):
    return list(padded_table(*args, **kwargs))


def test_padded_table__empty():
    assert _padded_table(()) == []


def test_padded_table__single_line():
    table = [(1, 20, 3000)]
    expected = ["1    20    3000"]
    assert expected == _padded_table(table)


def test_padded_table__two_lines():
    table = [(1, 20, 3000), (3000, 20, 1)]
    expected = ["1       20    3000", "3000    20    1"]
    assert expected == _padded_table(table)


def test_padded_table__three_lines():
    table = [(1, 20, 3000), (3000, 20, 1), (1, 2, 30)]
    expected = ["1       20    3000", "3000    20    1", "1       2     30"]
    assert expected == _padded_table(table)


_PT_COMMENT = "# An insightful comment goes here"
_PT_ROW_1 = (1, 20, 3000)
_PT_ROW_2 = (3000, 20, 1)
_PT_LINE_1 = "1       20    3000"
_PT_LINE_2 = "3000    20    1"

_PT_PERMUTATIONS = (
    ([_PT_COMMENT, _PT_ROW_1, _PT_ROW_2], [_PT_COMMENT, _PT_LINE_1, _PT_LINE_2]),
    ([_PT_ROW_1, _PT_COMMENT, _PT_ROW_2], [_PT_LINE_1, _PT_COMMENT, _PT_LINE_2]),
    ([_PT_ROW_1, _PT_ROW_2, _PT_COMMENT], [_PT_LINE_1, _PT_LINE_2, _PT_COMMENT]),
)


@pytest.mark.parametrize("table, expected", _PT_PERMUTATIONS)
def test_padded_table__with_text(table, expected):
    assert expected == _padded_table(table)


###############################################################################
###############################################################################
# Tests for 'parse_padded_table'


def _parse_padded_table(*args, **kwargs):
    return list(parse_padded_table(*args, **kwargs))


def test_parse_padded_table__empty():
    assert [] == _parse_padded_table([])


def test_parse_padded_table__header_only():
    assert [] == _parse_padded_table(["A  B  C  D"])


def test_parse_padded_table__single_row():
    table = ["A    B    C    D", "4    3    2    1"]
    expected = [{"A": "4", "B": "3", "C": "2", "D": "1"}]
    assert expected == _parse_padded_table(table)


def test_parse_padded_table__two_rows():
    table = ["A     B     C     D", "4     3     2     1", "AB    CD    EF    GH"]
    expected = [
        {"A": "4", "B": "3", "C": "2", "D": "1"},
        {"A": "AB", "B": "CD", "C": "EF", "D": "GH"},
    ]
    assert expected == _parse_padded_table(table)


# Any amount of whitespace is allowed
def test_parse_padded_table__single_row__with_whitespace():
    table = ["A   B    C       E F", "1        0  1    2   3"]
    expected = [{"A": "1", "B": "0", "C": "1", "E": "2", "F": "3"}]
    assert expected == _parse_padded_table(table)


# Other whitespace should be ignored
def test_parse_padded_table__single_row__with_tabs():
    table = ["A\t\t\t\tB", "1\t\t\t\t0"]
    expected = [{"A": "1", "B": "0"}]
    assert expected == _parse_padded_table(table)


_PT_COMMENTS_LINE_AND_EMPTY_1 = "A         B      C"
_PT_COMMENTS_LINE_AND_EMPTY_2 = "3000      20      1"

__PT_COMMENTS_LINE_AND_EMPTY = (
    ["# comment", _PT_COMMENTS_LINE_AND_EMPTY_1, _PT_COMMENTS_LINE_AND_EMPTY_2],
    [_PT_COMMENTS_LINE_AND_EMPTY_1, " # comment", _PT_COMMENTS_LINE_AND_EMPTY_2],
    [_PT_COMMENTS_LINE_AND_EMPTY_1, _PT_COMMENTS_LINE_AND_EMPTY_2, "  # comment"],
    ["", _PT_COMMENTS_LINE_AND_EMPTY_1, _PT_COMMENTS_LINE_AND_EMPTY_2],
    [_PT_COMMENTS_LINE_AND_EMPTY_1, "  ", _PT_COMMENTS_LINE_AND_EMPTY_2],
    [_PT_COMMENTS_LINE_AND_EMPTY_1, _PT_COMMENTS_LINE_AND_EMPTY_2, "   "],
)


@pytest.mark.parametrize("lines", __PT_COMMENTS_LINE_AND_EMPTY)
def test_padded_table__comments_and_empty_lines(lines):
    expected = [{"A": "3000", "B": "20", "C": "1"}]
    assert expected == _parse_padded_table(lines)


@pytest.mark.parametrize("postfix", ("\r", "\n", "\r\n"))
def test_padded_table__newlines(postfix):
    expected = [{"A": "3000", "B": "20", "C": "1"}]

    line_1 = "A         B       C" + postfix
    line_2 = "3000      20      1" + postfix
    assert expected == _parse_padded_table([line_1, line_2])


def test_padded_table__padding__comments__whitespace():
    expected = [{"A": "3000", "B": "20", "C": "1"}]
    lines = ["A         B       C", "3000      20      1", "  # useless comment"]
    assert expected == _parse_padded_table(lines)


def test_parse_padded_table__malformed_table_0():
    table = ["A    B    C    D", "4    3    2"]
    with pytest.raises(TableError):
        _parse_padded_table(table)


def test_parse_padded_table__malformed_table_1():
    table = ["A    B    C    D", "4    3    2    1    0"]
    with pytest.raises(TableError):
        _parse_padded_table(table)


###############################################################################
###############################################################################
# Tests for 'parse_linse'
def _this(*args):
    return args


def _parse_lines(*args, **kwargs):
    return list(parse_lines(*args, **kwargs))


def test_parse_lines__empty_file():
    def _assert_false():
        assert False  # pragma: no coverage

    assert _parse_lines([], _assert_false) == []


def test_parse_lines__single():
    assert _parse_lines(["abc line1 \n"], _this) == [("abc line1", 9)]


_PT_COMMENTS_AND_EMPTY_LINES = (
    ["# comment\n", "abc line1 \n", "def line2 \n"],
    ["abc line1 \n", " # comment\n", "def line2 \n"],
    ["abc line1 \n", "def line2 \n", "   # comment\n"],
    ["\n", "abc line1 \n", "def line2 \n"],
    ["abc line1 \n", " \n", "def line2 \n"],
    ["abc line1 \n", "def line2 \n", "   \n"],
)


@pytest.mark.parametrize("lines", _PT_COMMENTS_AND_EMPTY_LINES)
def test_parse_lines__comments_and_empty_lines(lines):
    expected = [("abc line1", 9), ("def line2", 9)]
    assert _parse_lines(lines, _this) == expected


@pytest.mark.parametrize("postfix", ("\r", "\n", "\r\n"))
def test_parse_lines__padding__newlines(postfix):
    expected = [("abc line1", 9), ("def line2", 9)]

    line_1 = "abc line1 " + postfix
    line_2 = "def line2 " + postfix
    assert expected == _parse_lines([line_1, line_2], _this)


def test_parse_lines__uncallable():
    with pytest.raises(TypeError):
        _parse_lines([], 1)


def parse_lines__binary():
    lines = [b"# foo", b"12", b"3456"]
    expected = [(12, 2), (3456, 4)]

    def parser(value, length):
        return (int(value), length)

    assert parse_lines(lines, parser) == expected


###############################################################################
###############################################################################
# Tests for 'parse_lines_by_contig'
_RecordMock = collections.namedtuple("_RecordMock", "contig value")


def test_parse_lines_by_contig__single_contig():
    lines = ["abc line1 \n", "abc line2 \n"]

    def _parse(line, length):
        assert len(line) == length
        return _RecordMock(*line.split())

    expected = {"abc": [_RecordMock("abc", "line1"), _RecordMock("abc", "line2")]}
    assert parse_lines_by_contig(lines, _parse) == expected


def test_parse_lines__two_contigs():
    lines = ["abc line1 \n", "def line2 \n"]

    def _parse(line, length):
        assert len(line) == length
        return _RecordMock(*line.split())

    expected = {
        "abc": [_RecordMock("abc", "line1")],
        "def": [_RecordMock("def", "line2")],
    }
    assert parse_lines_by_contig(lines, _parse) == expected



#!/usr/bin/python3
# -*- coding: utf-8 -*-
# Copyright (c) 2014 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import os
import logging
import sys

import pysam

import paleomix.common.argparse as argparse
import paleomix.pipelines.zonkey.database as database

from paleomix.common.formats.fasta import FASTA
from paleomix.common.formats.msa import MSA
from paleomix.common.formats.phylip import interleaved_phy
from paleomix.common.utilities import fragment


def majority_base(site):
    best, count = "N", 0
    for nuc in "ACGT":
        if site[nuc] > count:
            best = nuc
            count = site[nuc]
        elif site[nuc] == count:
            best = "N"

    return best


def majority_sequence(handle, padding, contig_name, contig_length):
    sequence = [dict.fromkeys("ACGTN", 0) for _ in range(contig_length)]

    for column in handle.pileup(contig_name):
        position = sequence[column.pos]

        for alignment in column.pileups:
            seq = alignment.alignment.seq
            pos = alignment.query_position

            if pos is not None:
                position[seq[pos]] += 1

    if padding:
        offset = len(sequence) - padding
        for idx in range(padding):
            dst = sequence[idx]
            src = sequence[idx + offset]

            for key, value in src.items():
                dst[key] += value

        del sequence[-padding:]

    covered = coverage = 0
    for counts in sequence:
        total = sum(counts.values()) - counts["N"]
        coverage += total

        if total:
            covered += 1

    statistics = {
        "sequence_len": len(sequence),
        "sequence_name": contig_name,
        "nucleotides": coverage,
        "covered_sites": covered,
        "covered_pct": round((100.0 * covered) / len(sequence), 1),
        "mean_coverage": round(coverage / float(len(sequence)), 1),
    }

    return statistics, "".join(map(majority_base, sequence))


def align_majority(reference, majority):
    aligned = []
    reference_iter = iter(reference).__next__

    for nucleotide in majority:
        reference = reference_iter()
        while reference == "-":
            reference = reference_iter()
            aligned.append("-")

        aligned.append(nucleotide)

    return "".join(aligned)


def truncate_sequences(sequences, name):
    result = {}
    to_len = len(sequences[name].sequence)
    for name, record in sequences.items():
        result[name] = FASTA(
            name=record.name, meta=record.meta, sequence=record.sequence[:to_len]
        )

    return result


def filter_sequences(sequences):
    selection = {}
    for key, record in sequences.items():
        if "EXCLUDE" not in map(str.strip, record.meta.upper().split(";")):
            selection[key] = record

    return selection


def sequences_to_msa(sequences):
    records = []
    for name, record in sorted(sequences.items()):
        records.append(record)

    return MSA(records)


def parse_args(argv):
    parser = argparse.ArgumentParser(prog="paleomix zonkey:mito")
    parser.add_argument("database")
    parser.add_argument("bam")
    parser.add_argument("output_prefix")

    return parser.parse_args(argv)


def main(argv):
    args = parse_args(argv)
    data = database.ZonkeyDB(args.database)
    sequences = data.mitochondria
    log = logging.getLogger(__name__)

    try:
        handle = pysam.AlignmentFile(args.bam)
    except (IOError, ValueError) as error:
        log.error("Error reading BAM file: %s", error)
        return 1

    with handle:
        bam_info = data.validate_bam_handle(handle)
        if bam_info is None:
            return 1
        elif not bam_info.is_mitochondrial:
            log.error("BAM does not contain any known mitochondrial sequence")
            return 1

        reference = sequences[bam_info.mt_contig]
        stats, majority = majority_sequence(
            handle,
            padding=bam_info.mt_padding,
            contig_name=bam_info.mt_contig,
            contig_length=bam_info.mt_length,
        )

        sequences["Sample"] = FASTA(
            name="Sample",
            meta=None,
            sequence=align_majority(reference.sequence, majority),
        )

        # Truncate all sequences to match the (now) unpadded sample sequence
        sequences = truncate_sequences(sequences, "Sample")

        sequences = filter_sequences(sequences)

    with open(args.output_prefix + ".summary", "w") as handle:
        stats["filename"] = os.path.abspath(args.bam)

        for key, value in sorted(stats.items()):
            handle.write("{}: {}\n".format(key, value))

    with open(args.output_prefix + ".phy", "w") as handle:
        handle.write(interleaved_phy(sequences_to_msa(sequences)))

    with open(args.output_prefix + ".fasta", "w") as handle:
        for key, record in sorted(sequences.items()):
            handle.write(">{}\n".format(key))
            for line in fragment(60, record.sequence):
                handle.write("{}\n".format(line))

        return 0


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



from .reads import Reads
from .lane import Lane
from .library import Library
from .sample import Sample
from .prefix import Prefix
from .target import Target
from .statistics import add_statistics_nodes


__all__ = [
    "Reads",
    "Lane",
    "Library",
    "Sample",
    "Prefix",
    "Target",
    "add_statistics_nodes",
]
